/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi@cisco.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi.c,v 1.23 2001/10/14 22:10:05 smferris Exp $ 
 *
 */

/* there's got to be a better way to wait for child processes created by kernel_thread */
static int errno = 0;
#define __KERNEL_SYSCALLS__

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>

#include <linux/sched.h>
#include <asm/io.h>
#include <asm/byteorder.h>
#include <linux/stddef.h>
#include <linux/version.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/file.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/blk.h>
#include <linux/types.h>
#include <linux/stat.h>
#include <linux/malloc.h>
#include <linux/config.h>
#include <linux/poll.h>
#include <linux/smp_lock.h>
#include <linux/slab.h>
#include <linux/kernel.h>
#include <linux/wait.h>
#include <linux/net.h>
#include <net/sock.h>
#include <linux/socket.h>
#include <linux/errno.h>
#include <linux/unistd.h>

#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
#include <asm/semaphore.h>
#else
#include <asm/spinlock.h>
#endif
#include <asm/uaccess.h>
#include <scsi/sg.h>
#include <endian.h>

#include "sd.h"
#include "scsi.h"
#include "hosts.h"


/* #define DEBUG_INIT   1 */
/* #define DEBUG_QUEUE  1 */
/* #define DEBUG_FLOW   1 */
#define DEBUG_ERROR 1
/* #define DEBUG_TRACE  1  */

#define TEST_ABORTS 0
#define ABORT_FREQUENCY 1000

/* requires TEST_ABORTS 1 */
#define TEST_RESETS 0
#define RESET_FREQUENCY 100


#include "iscsi-common.h"
#include "iscsi-protocol.h"
#include "iscsi-ioctl.h"
#include "iscsi-trace.h"
#include "iscsi.h"


/*
 *  IMPORTANT NOTE: to prevent deadlock, when holding multiple locks,
 *  the following locking order must be followed at all times:
 *
 *  io_request_lock         - mid-layer acquires before calling queuecommand, eh_*, we must acquire before done() callback
 *  hba_list_lock           - acceess to collection of HBA instances
 *  hba->session_lock       - access to an HBA's collection of sessions   
 *  session->scsi_cmnd_lock - access to a session's list of Scsi_Cmnds
 *  session->task_lock      - access to a session's collections of tasks
 *  hba->task_lock          - for task alloc/free 
 */

MODULE_AUTHOR("Cisco Systems, Inc.");
MODULE_DESCRIPTION("iSCSI Driver");
MODULE_LICENSE("GPL");


/* useful 2.4-ism */
#ifndef set_current_state
#define set_current_state(state_value) do { current->state = state_value; mb(); } while(0)
#endif


/* we'd prefer to do all the locking ourselves, but the SCSI mid-layer 
 * tends to call us with the io_request_lock held, and requires that we
 * get the lock before calling a SCSI command's done() callback.
 * This is supposed to be removed in lk 2.5, so make it conditional at compile-time.
 */

#define MIDLAYER_USES_IO_REQUEST_LOCK

#ifdef MIDLAYER_USES_IO_REQUEST_LOCK

# if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
typedef unsigned long cpu_flags_t;
# else
typedef unsigned int cpu_flags_t;
# endif

# define RELEASE_IO_REQUEST_LOCK \
    { \
        cpu_flags_t flags; \
        save_flags(flags); \
        spin_unlock_irqrestore(&io_request_lock, flags); \
        sti(); /* enable interrupts */ \
    }

# define REACQUIRE_IO_REQUEST_LOCK \
    { \
        cpu_flags_t flags; \
        spin_lock_irqsave(&io_request_lock, flags); \
    }

# define DECLARE_CPU_FLAGS cpu_flags_t cpu_flags
# define LOCK_IO_REQUEST_LOCK spin_lock_irqsave(&io_request_lock, cpu_flags);
# define UNLOCK_IO_REQUEST_LOCK spin_unlock_irqrestore(&io_request_lock, cpu_flags);

#else
# define RELEASE_IO_REQUEST_LOCK
# define REACQUIRE_IO_REQUEST_LOCK
# define DECLARE_CPU_FLAGS 
# define LOCK_IO_REQUEST_LOCK
# define UNLOCK_IO_REQUEST_LOCK
#endif

/* Scsi_cmnd->result */
#define DRIVER_BYTE(byte)   ((byte) << 24)
#define HOST_BYTE(byte)     ((byte) << 16) /* HBA codes */
#define MSG_BYTE(byte)      ((byte) << 8)
#define STATUS_BYTE(byte)   ((byte))  /* SCSI status */


static int ControlOpen( struct inode *inode, struct file *file );
static int ControlClose( struct inode *inode, struct file *file );
static int ControlIoctl( struct inode *inode,
                         struct file *file,
                         unsigned int cmd,
                         unsigned long arg );

static int ControlMajor;
static const char *ControlName = "iscsictl";

#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
static struct file_operations ControlFops = {
    owner: THIS_MODULE,
    ioctl: ControlIoctl,    /* ioctl */
    open: ControlOpen,      /* open */
    release: ControlClose,  /* release */
};
#else
static struct file_operations ControlFops = {
    NULL,                   /* lseek */
    NULL,                   /* read */
    NULL,                   /* write */
    NULL,                   /* readdir */
    NULL,                   /* poll */
    ControlIoctl,       /* ioctl */
    NULL,                   /* mmap */
    ControlOpen,        /* open */
    NULL,                   /* flush */
    ControlClose,       /* release */
};
#endif

spinlock_t hba_list_lock = SPIN_LOCK_UNLOCKED;
static iscsi_hba_t *hba_list = NULL;

static unsigned int init_module_complete = 0;
static volatile int TimerActive = 0;
static atomic_t NumTargets = ATOMIC_INIT(0);


#ifdef DEBUG_TRACE
spinlock_t TraceSpinLock = SPIN_LOCK_UNLOCKED;
#endif

#ifdef DEBUG_TRACE
static iscsiTraceEntry TraceTable[ISCSI_TRACE_COUNT];
static int TraceIndex=0;
#define ISCSI_TRACE( P_FLG, P_LUN, P_CMD, P_TGT, P_ITT, P_DAT ) \
    iscsi_fill_trace( P_FLG, P_LUN, P_CMD, P_TGT, P_ITT, P_DAT );
#else
#define ISCSI_TRACE( P_FLG, P_LUN, P_CMD, P_TGT, P_ITT, P_DAT )
#endif

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,27) )
/* note change modeled per linux2.4 drivers/scsi/ips.c */
struct proc_dir_entry proc_dir_iscsi = {
#ifdef PROC_SCSI_ISCSI
    PROC_SCSI_ISCSI,
#else
    PROC_SCSI_NOT_PRESENT,
#endif
    5,
    "iscsi",
    S_IFDIR|S_IRUGO|S_IXUGO,
    2
};
#endif

/* wake up the tx_thread without ever losing the wakeup event */
static void wake_tx_thread(int control_bit, iscsi_session_t *session)
{
    /* tell the Tx thread what to do when it wakes up. */
    set_bit(control_bit, &session->control_bits);
    
    /* We make a condition variable out of a wait queue and atomic test&clear.
     * May get spurious wake-ups, but no wakeups will be lost.
     * this is cv_signal().  wait_event_interruptible is cv_wait().
     */
    set_bit(TX_WAKE, &session->control_bits);
    wake_up_interruptible(&session->tx_wait_q);
}


/* compare against 2^31 */
#define SNA32_CHECK 2147483648UL

/* Serial Number Arithmetic, 32 bits, less than, RFC1982 */
static int sna_lt(uint32_t n1, uint32_t n2)
{
    return ((n1 != n2) && 
            (((n1 < n2) && ((n2 - n1) < SNA32_CHECK)) || ((n1 > n2) && ((n2 - n1) < SNA32_CHECK))));
}

/* Serial Number Arithmetic, 32 bits, less than, RFC1982 */
static int sna_lte(uint32_t n1, uint32_t n2)
{
    return ((n1 == n2) ||
            (((n1 < n2) && ((n2 - n1) < SNA32_CHECK)) || ((n1 > n2) && ((n2 - n1) < SNA32_CHECK))));
}

/* do a similar wrapping comparision for jiffies, which are unsigned longs */
#define JIFFY_CHECK (1UL << ((sizeof(unsigned long) * 8) - 1))
static int jiffy_lt(unsigned long n1, unsigned long n2)
{
    return ((n1 != n2) && 
            (((n1 < n2) && ((n2 - n1) < JIFFY_CHECK)) || ((n1 > n2) && ((n2 - n1) < SNA32_CHECK))));
}

/* possibly update the ExpCmdSN and MaxCmdSN */
static void updateSN(iscsi_session_t *session, UINT32 expcmdsn, UINT32 maxcmdsn) 
{
    /* standard specifies this check for when to update expected and max sequence numbers */
    if (!sna_lt(maxcmdsn, expcmdsn - 1)) {
        if (! sna_lt(expcmdsn, session->ExpCmdSn)) {
            session->ExpCmdSn = expcmdsn;
        }
        if (! sna_lt(maxcmdsn, session->MaxCmdSn)) {
            session->MaxCmdSn = maxcmdsn;
            /* wake the tx thread to try sending more commands */
            wake_tx_thread(TX_SCSI_COMMAND, session);
        }

        /* record whether or not the command window for this session has closed,
         * so that we can ping the target periodically to ensure we eventaully
         * find out that the window has re-opened.  
         */
        if (maxcmdsn == expcmdsn - 1)
            set_bit(SESSION_WINDOW_CLOSED, &session->control_bits);
        else
            clear_bit(SESSION_WINDOW_CLOSED, &session->control_bits);

        DEBUG_FLOW3("iSCSI: %s - ExpCmdSN %u, MaxCmdSN %u\n", 
                    session->TargetName, session->ExpCmdSn, session->MaxCmdSn);
    }
}

static iscsi_session_t *find_session_for_cmnd(Scsi_Cmnd *sc)
{
    iscsi_session_t *session = NULL;
    iscsi_hba_t *hba;
    cpu_flags_t cpu_flags;

    if (!sc->host)
        return NULL;

    if (!sc->host->hostdata)
        return NULL;

    hba = (iscsi_hba_t *)sc->host->hostdata;

    /* find the session for this command */
    spin_lock_irqsave(&hba->session_lock, cpu_flags);
    session = hba->session_list_head;
    while (session && (session->channel != sc->channel || session->target_id != sc->target))
        session = session->next;
    spin_unlock_irqrestore(&hba->session_lock, cpu_flags);

    return session;
}

static iscsi_task_t *alloc_task(iscsi_session_t *session )
{
    iscsi_hba_t *hba = session->hba;
    iscsi_task_t *task = NULL;

    /* get one from the HBA's free list */
    spin_lock( &hba->task_lock);
    if ((task = hba->free_task_head)) {
        hba->free_task_head = task->next;
        hba->active_tasks++;

        memset(task, 0, sizeof(iscsi_task_t) );
        task->ttt = RSVD_TASK_TAG;
        task->mgmt_itt = RSVD_TASK_TAG;
        task->next = task->prev = NULL;
        task->order_next = task->order_prev = NULL;
    }
    /* FIXME: dynamically allocate tasks */
    spin_unlock(&hba->task_lock);

    return task;
}


static void free_task( iscsi_task_t *task )
{
    iscsi_session_t *session = task->session;
    iscsi_hba_t *hba;

    if (! task) {
        DEBUG_ERR0("iSCSI: free_task couldn't free NULL task\n");
        return;
    }
    if (! session) {
        DEBUG_ERR1("iSCSI: free_task couldn't find session for task %p\n", task);
        return;
    }
    hba = session->hba;
    if ( ! hba ) {
        DEBUG_ERR1("iSCSI: free_task couldn't find hba for task %p\n", task);
        return;
    }

    DEBUG_QUEUE1("iSCSI: free_task %p\n", task);

    /* disconnect the Scsi_Cmnd and the task */
    if (task->scsi_cmnd && (iscsi_task_t *)task->scsi_cmnd->host_scribble == task)
        task->scsi_cmnd->host_scribble = NULL;
    task->scsi_cmnd = NULL;

    /* zero out the task settings */
    task->session = NULL;
    task->itt = RSVD_TASK_TAG;
    task->mgmt_itt = RSVD_TASK_TAG;
    task->next = task->prev = NULL;
    task->order_next = task->order_prev = NULL;
    atomic_set(&task->refcount, 0);

    /* put the task on the HBA's free list */
    spin_lock( &hba->task_lock);
    if ( hba->free_task_head )
        hba->free_task_tail->next = task;
    else
        hba->free_task_head = task;
    hba->free_task_tail = task;
    hba->active_tasks--;
    spin_unlock( &hba->task_lock);

    /* wake the tx thread after freeing the task, in case the tx
     * thread blocked after running out of available tasks.
     */
    wake_tx_thread(TX_SCSI_COMMAND, session);
}


/* must hold the task_lock to call this */
static iscsi_task_t *find_task(iscsi_task_collection_t *collection, uint32_t itt)
{
    iscsi_task_t *task = collection->head;
    
    while (task) {
        if (task->itt == itt) {
            DEBUG_FLOW3("iSCSI: found itt %u, task %p, refcount %d\n", itt, task, atomic_read(&task->refcount));
            return task;
        }
        task = task->next;
    }

    return NULL;
}

#if 0
/* don't actually use this at the moment */
/* must hold the task_lock to call this */
static iscsi_task_t *find_mgmt_task(iscsi_task_collection_t *collection, uint32_t mgmt_itt)
{
    iscsi_task_t *task = collection->head;
    
    while (task) {
        if (task->mgmt_itt == mgmt_itt) {
            DEBUG_FLOW2("iSCSI: found mgmt_itt %u, task %p\n", mgmt_itt, task);
            return task;
        }
        task = task->next;
    }

    return NULL;
}
#endif

#if 0
/* don't actually need this at the moment */
/* must hold the task_lock to call this */
static iscsi_task_t *find_task_for_cmnd(iscsi_task_collection_t *collection, Scsi_Cmnd *sc)
{
    iscsi_task_t *task = collection->head;
    
    while (task) {
        if (task->scsi_cmnd == sc) {
            DEBUG_FLOW3("iSCSI: found itt %u, task %p for cmnd %p\n", task->itt, task, sc);
            return task;
        }
        task = task->next;
    }

    return NULL;
}
#endif

/* add a task to the collection.  Must hold the task_lock to do this. */
static void add_task(iscsi_task_collection_t *collection, iscsi_task_t *task)
{
    if (task->prev || task->next)
        printk("iSCSI: bug, adding task %p, prev %p, next %p, to collection %p\n",
               task, task->prev, task->next, collection);

    if (collection->head) {
        task->next = NULL;
        task->prev = collection->tail;
        collection->tail->next = task;
        collection->tail = task;
    }
    else {
        collection->head = collection->tail = task;
        task->prev = task->next = NULL;
    }
}

/* must hold the task_lock when calling this */
static iscsi_task_t *pop_task(iscsi_task_collection_t *collection)
{
    iscsi_task_t *task = NULL;

    if ((task = collection->head)) {
        /* pop the head */
        if ((collection->head = task->next))
            collection->head->prev = NULL;
        else
            collection->tail = NULL;

        /* and return it */
        task->prev = NULL;
        task->next = NULL;
        
        return task;
    }

    return NULL;
}

/* must hold the task_lock when calling this */
static void push_task(iscsi_task_collection_t *collection, iscsi_task_t *task)
{
    if (task) {
        task->prev = NULL;
        task->next = collection->head;
        if (collection->head)
            collection->head->prev = task;
        collection->head = task;
    }
}


/* if the task for the itt is found in the collection, remove it, and return it.
 * otherwise, return NULL.  Must hold the task_lock to call this.
 */
static iscsi_task_t *remove_task(iscsi_task_collection_t *collection, uint32_t itt)
{
    iscsi_task_t *task = NULL;
    iscsi_task_t *search = collection->head;

    while (search) {
        if (search->itt == itt) {
            task = search;
            /* remove it */
            if (task == collection->head) {
                if ((collection->head = task->next))
                    collection->head->prev = NULL;
                else
                    collection->tail = NULL;
            }
            else if (task == collection->tail) {
                collection->tail = task->prev;
                collection->tail->next = NULL;
            }
            else {
                task->next->prev = task->prev;
                task->prev->next = task->next;
            }
            task->next = NULL;
            task->prev = NULL;
            return task;
        }
        search = search->next;
    }

    return NULL;
}

/* if the task for the mgmt_itt is found in the collection, remove it, and return it.
 * otherwise, return NULL.  Must hold the task_lock to call this.
 */
static iscsi_task_t *remove_mgmt_task(iscsi_task_collection_t *collection, uint32_t mgmt_itt)
{
    iscsi_task_t *task = NULL;
    iscsi_task_t *search = collection->head;

    while (search) {
        if (search->mgmt_itt == mgmt_itt) {
            task = search;
            /* remove it */
            if (task == collection->head) {
                if ((collection->head = task->next))
                    collection->head->prev = NULL;
                else
                    collection->tail = NULL;
            }
            else if (task == collection->tail) {
                collection->tail = task->prev;
                collection->tail->next = NULL;
            }
            else {
                task->next->prev = task->prev;
                task->prev->next = task->next;
            }
            task->next = NULL;
            task->prev = NULL;
            return task;
        }
        search = search->next;
    }

    return NULL;
}

/* if the task for the itt is found in the collection, remove it, and return it.
 * otherwise, return NULL.  Must hold the task_lock to call this.
 */
static iscsi_task_t *remove_task_for_cmnd(iscsi_task_collection_t *collection, Scsi_Cmnd *sc)
{
    iscsi_task_t *task = NULL;
    iscsi_task_t *search = collection->head;

    while (search) {
        if (search->scsi_cmnd == sc) {
            task = search;
            /* remove it */
            if (task == collection->head) {
                if ((collection->head = task->next))
                    collection->head->prev = NULL;
                else
                    collection->tail = NULL;
            }
            else if (task == collection->tail) {
                collection->tail = task->prev;
                collection->tail->next = NULL;
            }
            else {
                task->next->prev = task->prev;
                task->prev->next = task->next;
            }
            task->next = NULL;
            task->prev = NULL;
            return task;
        }
        search = search->next;
    }

    return NULL;
}


/* allocate an itt, and add the task to the session's task FIFO */
static void add_task_to_session(iscsi_session_t *session, iscsi_task_t *task)
{
    spin_lock(&session->task_lock);

    if (atomic_read(&session->num_tasks) == 0) {
        /* session going from idle to active, pretend we just
         * received something, so that the idle period before this doesn't
         * cause an immediate timeout.
         */
        session->last_rx = jiffies;
    }
    atomic_inc(&session->num_tasks);

    /* set task info */
    task->session = session;

    /* allocate the ITT */
    task->itt = session->itt++;
    /* iSCSI reserves 0xFFFFFFFF, this driver reserves 0 */
    if (session->itt == RSVD_TASK_TAG)
        session->itt = 1;
    DEBUG_QUEUE5("iSCSI: task %p allocated itt %u for command %p, session %p to %s\n", 
                 task, task->itt, task->scsi_cmnd, session, session->TargetName);

    /* add it to the session task ordering list */
    if (session->arrival_order.head) {
        session->arrival_order.tail->order_next = task;
        task->order_prev = session->arrival_order.tail;
        task->order_next = NULL;
        session->arrival_order.tail = task;
    }
    else {
        session->arrival_order.head = session->arrival_order.tail = task;
        task->order_prev = NULL;
        task->order_next = NULL;
    }

    /* add it to the rx_tasks as well */
    add_task(&session->rx_tasks, task);

    DEBUG_FLOW4("iSCSI: task %p, itt %u, added to session %p to %s\n", task, task->itt, session, session->TargetName);
    spin_unlock(&session->task_lock);
}

static void remove_session_task(iscsi_session_t *session, iscsi_task_t *task)
{
   /* remove the task from the session's arrival_order collection */
    if (task == session->arrival_order.head) {
        if ((session->arrival_order.head = task->order_next))
            session->arrival_order.head->order_prev = NULL;
        else
            session->arrival_order.tail = NULL;
    }
    else if (task == session->arrival_order.tail) {
        session->arrival_order.tail = task->order_prev;
        session->arrival_order.tail->order_next = NULL;
    }
    else {
        /* we should always be in the middle, 
         * but check pointers to make sure we don't crash the kernel 
         * if the function is called for task not in the session.
         */
        if (task->order_next && task->order_prev) {
            task->order_next->order_prev = task->order_prev;
            task->order_prev->order_next = task->order_next;
        }
        else {
            DEBUG_FLOW3("iSCSI: can't remove task %p from session %p to %s, not in session\n", task, session, session->TargetName);
            spin_unlock(&session->task_lock);
            return;
        }
    }
    task->order_prev = NULL;
    task->order_next = NULL;
}

static void remove_task_from_session(iscsi_session_t *session, iscsi_task_t *task)
{
    DEBUG_FLOW4("iSCSI: removing itt %u, task %p from session %p to %s\n", 
                task->itt, task, session, session->TargetName);
    spin_lock(&session->task_lock);

    /* remove it from the session's task ordering collection */
    remove_session_task(session, task);
 
    /* decrement the session's task count (used to decide timeouts) */
    atomic_dec(&session->num_tasks);
    DEBUG_QUEUE4("iSCSI: removed itt %u, task %p from session %p to %s\n", 
                 task->itt, task, session, session->TargetName);

    spin_unlock(&session->task_lock);
}

static int iscsi_recvmsg( iscsi_session_t *session, struct msghdr *msg, int len )
{
    int rc;
    mm_segment_t oldfs;

    oldfs = get_fs();
    set_fs( get_ds() );
    
    /* FIXME: ought to loop handling short reads, unless a signal occurs */
    rc = sock_recvmsg( session->socket, msg, len, MSG_WAITALL);
    if (rc > 0) {
        session->last_rx = jiffies;
        mb();
    }

    set_fs( oldfs );
    return rc;
}

static int iscsi_sendmsg( iscsi_session_t *session, struct msghdr *msg, int len )
{
    int rc;
    mm_segment_t oldfs;

    current->flags |= PF_MEMALLOC; /* FIXME: what is this, and do we need it? */
    oldfs = get_fs();
    set_fs( get_ds() );

    /* FIXME: ought to loop handling short reads, unless a signal occurs */
    rc = sock_sendmsg( session->socket, msg, len );

    set_fs( oldfs );
    current->flags &= ~PF_MEMALLOC;

    return rc;
}


#ifdef DEBUG_TRACE
static void
iscsi_fill_trace( unsigned char flag,
                    unsigned char lun,
                    unsigned char cmd,
                    unsigned char target,
                    unsigned int itt,
                    unsigned int data )
{
    iscsiTraceEntry *te;

    spin_lock( &TraceSpinLock);
    te = &TraceTable[TraceIndex];
    TraceIndex += 1;
    if ( TraceIndex >= ISCSI_TRACE_COUNT ) {
        TraceIndex = 0;
    }
    spin_unlock(&TraceSpinLock);

    te->lun = lun;
    te->cmd = cmd;
    te->itt = itt;
    te->flag = flag;
    te->data = data;
    te->target = target;
    te->activecmds = hba_list->active_tasks;
    te->jiffies = jiffies;

    return;
}
#endif

static int
iscsi_cmnd_direction( unsigned char cmd )
{
    switch (cmd) {
        case TEST_UNIT_READY:
        case START_STOP:
        case REZERO_UNIT:
        case WRITE_FILEMARKS:
        case SPACE:
        case ERASE:
        case ALLOW_MEDIUM_REMOVAL:
            /* just control commands */
            return ISCSI_CMD_CONTROL;
        case WRITE_6:           case WRITE_10:          case WRITE_12:
        case WRITE_LONG:        case WRITE_SAME:        case WRITE_BUFFER:
        case WRITE_VERIFY:      case WRITE_VERIFY_12:
        case COMPARE:           case COPY:              case COPY_VERIFY:
        case SEARCH_EQUAL:      case SEARCH_HIGH:       case SEARCH_LOW:
        case SEARCH_EQUAL_12:   case SEARCH_HIGH_12:    case SEARCH_LOW_12:
        case FORMAT_UNIT:       case REASSIGN_BLOCKS:   case RESERVE:
        case MODE_SELECT:       case MODE_SELECT_10:    case LOG_SELECT:
        case SEND_DIAGNOSTIC:   case CHANGE_DEFINITION: case UPDATE_BLOCK:
        case SET_WINDOW:        case MEDIUM_SCAN:       case SEND_VOLUME_TAG:
        case 0xea:
            return ISCSI_CMD_WRITE;
        default:
            return ISCSI_CMD_READ;
    }
    
    return -1;
}

/* tagged queueing */
static void
iscsi_set_tag( Scsi_Cmnd *cmd, struct IscsiScsiCmdHdr *hdr )
{
    if ( cmd->device->tagged_supported ) {
        switch( cmd->tag ) {
            case HEAD_OF_QUEUE_TAG:
                hdr->flags.attr = ISCSI_ATTR_HEAD_OF_QUEUE;
                break;
            case ORDERED_QUEUE_TAG:
                hdr->flags.attr = ISCSI_ATTR_ORDERED;
                break;
            default:
                hdr->flags.attr = ISCSI_ATTR_SIMPLE;
                break;
        }
    }
    else
        hdr->flags.attr = ISCSI_ATTR_UNTAGGED;
}



static void
iscsi_xmit_data(iscsi_task_t *task, uint32_t ttt, uint32_t data_offset, uint32_t data_length)
{
    struct msghdr msg;
    struct IscsiDataHdr stdh;
    Scsi_Cmnd *sc = NULL;
    iscsi_session_t *session = task->session;
    struct scatterlist *sglist = NULL;
    int wlen, rc, index, iovn;
    uint32_t segOffset=0;
    int i, remain, xfrlen, segSN = 0;
    int bytes_to_fill, bytes_from_segment;
    
    sc = task->scsi_cmnd;
    remain = data_length;
    
    memset( &stdh, 0, sizeof(stdh) );
    stdh.opcode = ISCSI_OP_SCSI_DATA;
    stdh.itt = htonl(task->itt);
    stdh.ttt = ttt;
    stdh.offset = htonl(data_offset);
    
    session->TxIov[0].iov_base = &stdh;
    session->TxIov[0].iov_len  = sizeof(stdh);

#ifdef DEBUG_FLOW
    printk("iSCSI: xmit_data for itt %u, task %p, bufflen %d, credit %d @ %u\n"
           "       buffer %p len %u, request_buf %p len %u\n",
           task->itt, task, sc->bufflen, remain, data_offset,
           sc->buffer, sc->bufflen, sc->request_buffer, sc->request_bufflen);
#endif
    
    /* Find the segment and offset within the segment to start writing from.
     * We use the offset to find it (amg)
     */
    index = -1;
    if ( sc->use_sg ) {
        sglist = (struct scatterlist *)sc->buffer;
        segOffset = data_offset;
        for (i = 0; i < sc->use_sg; i++) {
            if (segOffset < sglist->length) {
                index = i;
                break;
            }
            segOffset -= sglist->length;
            sglist++;
        }

        /* FIXME: we seem to be getting commands that indicate scatter-gather, 
         * but have all zeroes in the sglist.
         */
        if (index < 0) {
            sglist = (struct scatterlist *)sc->buffer;
            printk("iSCSI: xmit_data for itt %u couldn't find offset %u in sglist %p, sc %p, bufflen %u, use_sg %u\n",
                   task->itt, data_offset, sglist, sc, sc->bufflen, sc->use_sg);
            if (sc->buffer != sc->request_buffer) {
                printk("iSCSI: request_buffer %p len %u\n", sc->request_buffer, sc->request_bufflen);
            }
            for (i = 0; i < sc->use_sg; i++) {
                printk("iSCSI: buffer %p[%02d] = addr %p, len %u\n", (struct scatterlist *)sc->buffer, i, sglist->address, sglist->length);
                sglist++;
            }
            return;
        }

        DEBUG_FLOW4("iSCSI: index %d, sglist %p length %d, segOffset %d\n", index, sglist, sglist->length, segOffset);
    }

   
    /* Our starting point is now segOffset within segment index.
     * Start sending the data 
     */
    while (!signal_pending(current) && remain) {
        stdh.datasn = htonl(segSN);
        stdh.offset = htonl(data_offset);
        
        if (session->dataPDULength && (remain > session->dataPDULength)) {
            /* enforce a DataPDULength limit */
            bytes_to_fill = session->dataPDULength;
        } 
        else {
            bytes_to_fill = remain;
            stdh.final = 1;
            /* FIXME: why only pass ExpStatSN on the final Data PDU? */
            stdh.expstatsn = htonl(session->ExpStatSn);
        }
        
        /*        DEBUG_FLOW3("iSCSI: remain %d, bytes_to_fill %d, sc->use_sg %u\n", remain, bytes_to_fill, sc->use_sg); */

        if ( sc->use_sg ) {
            iovn = 1;
            
            xfrlen = 0;
            /* while there is more data and we want to send more data */
            while (bytes_to_fill > 0) {
                if (index >= sc->use_sg) {
                    printk("iSCSI: xmit_data index %d, sc->use_sg %d, out of buffer\n", index, sc->use_sg);
                    return;
                }
                if (signal_pending(current)) {
                    DEBUG_FLOW0("iSCSI: signal pending, returning from xmit_data\n");
                    return;
                }

                bytes_from_segment = sglist->length - segOffset;
                if ( bytes_from_segment > bytes_to_fill ) {
                    /* last data for this PDU */
                    xfrlen += bytes_to_fill;
                    session->TxIov[iovn].iov_base = sglist->address + segOffset;
                    session->TxIov[iovn].iov_len  = bytes_to_fill;
                    iovn++;
                    segOffset += bytes_to_fill;
                    DEBUG_FLOW3("iSCSI: index %d, xfrlen %d, to_fill %d, last segment\n", 
                                index, xfrlen, bytes_to_fill);
                    break;
                } 
                else {
                    /* need all of this segment, and more from the next */
                    xfrlen += bytes_from_segment;
                    session->TxIov[iovn].iov_base = sglist->address + segOffset;
                    session->TxIov[iovn].iov_len  = bytes_from_segment;
                    bytes_to_fill -= bytes_from_segment;
                    iovn += 1;
                    index += 1;
                    sglist += 1;
                    segOffset = 0;
                }
#ifdef DEBUG
                DEBUG_FLOW5("iSCSI: index %d, xfrlen %d, to_fill %d, from_segment %d, sglist %p\n", 
                            index, xfrlen, bytes_to_fill, bytes_from_segment, sglist);
#endif
            }
            
            if (xfrlen <= 0) {
                printk("iSCSI: Error xmit_data picked xfrlen of 0, index %d, sc->use_sg %d, bytes_to_fill %d\n",
                       index, sc->use_sg, bytes_to_fill);
                /* close the connection */
                if (session->rx_pid)
                    kill_proc(session->rx_pid, SIGPIPE, 1);
                return;
            }
        } 
        else {
            /* no scatter-gather */
            if ((sc->buffer + data_offset + bytes_to_fill) <= (sc->buffer + sc->bufflen)) {
                /* send all the data */
                session->TxIov[1].iov_base = sc->buffer + data_offset;
                session->TxIov[1].iov_len = xfrlen = bytes_to_fill;
                iovn = 2;
            }
            else if ((sc->buffer + data_offset) < (sc->buffer + sc->bufflen)) {
                /* send some data, but can't send all requested */
                xfrlen = sc->bufflen - data_offset;
                printk("iSCSI: xmit_data ran out of data, buffer %p len %u but offset %d length %d, sending final %d bytes\n",
                       sc->buffer, sc->bufflen, data_offset, bytes_to_fill, xfrlen);
                session->TxIov[1].iov_base = sc->buffer + data_offset;
                session->TxIov[1].iov_len = xfrlen;
                iovn = 2;
                stdh.final = 1;
                remain = xfrlen;
            }
            else {
                /* can't send any data */
                printk("iSCSI: xmit_data ran out of data, buffer %p len %u but offset %d length %d, sending no more data\n",
                       sc->buffer, sc->bufflen, data_offset, bytes_to_fill);
                return;
            }
        }

        hton24(stdh.dlength, xfrlen);
        
        memset( &msg, 0, sizeof(msg) );
        msg.msg_iov = &session->TxIov[0];
        msg.msg_iovlen = iovn;
        /* msg.msg_flags = MSG_DONTWAIT; */

        /* FIXME: can we really hold the largest possible size we might need to send in an int? */
        wlen = sizeof(stdh) + xfrlen;
        rc = iscsi_sendmsg( session, &msg, wlen );
        if ( rc != wlen ) {
            DEBUG_ERR2("iSCSI: xmit_data %d, error %d\n",wlen,rc);
            /* close the connection */
            if (session->rx_pid)
                kill_proc(session->rx_pid, SIGPIPE, 1);
            return;
        }
        
        remain  -= xfrlen;

        DEBUG_FLOW5("iSCSI: xmit_data sent %d @ %u for itt %u, remaining %d, final %d\n",
                    xfrlen, data_offset, task->itt, remain, stdh.final);

        data_offset += xfrlen;

        /* FIXME: ??? */
        if (++segSN == 0xffffffff)
            segSN++;
    }
    
    ISCSI_TRACE( ISCSI_TRACE_TxData, sc->lun, sc->cmnd[0], sc->target, (unsigned int)task, ttt );
    
    return;
}

static int iscsi_xmit_task( iscsi_task_t *task)
{
    struct msghdr msg;
    struct iovec iov;
    struct IscsiScsiCmdHdr stsch;
    int rc, wlen, directx;
    iscsi_session_t *session = task->session;
    Scsi_Cmnd *sc = task->scsi_cmnd;

#ifdef DEBUG_FLOW
    printk("iSCSI: xmit_task %p, itt %u to (%u %u %u %u), Cmd 0x%x, cmd_len %u, bufflen %u, request_len %u\n",  
           task, task->itt, sc->host->host_no, sc->channel, sc->target,sc->lun, sc->cmnd[0], 
           sc->cmd_len, sc->bufflen, sc->request_bufflen);
#endif

    wlen = sizeof(stsch);
    memset( &stsch, 0, sizeof(stsch) );

    directx = iscsi_cmnd_direction( sc->cmnd[0] );
    if ( directx < 0 ) {
        printk("iSCSI: xmit_task - unsupported SCSI command, 0x%x\n",sc->cmnd[0]);
        return 0;
    }

    task->flags |= directx;
    task->cmdsn = session->CmdSn;

    if ( directx != ISCSI_CMD_CONTROL ) {
        /* read or write type command */
        if ( directx == ISCSI_CMD_READ ) {
            /* read */
            stsch.flags.read_data = 1;
        }
        else {
            /* write */
            stsch.flags.write_data = 1;                                        
        }
        stsch.data_length = htonl(sc->bufflen);
    }

    iscsi_set_tag( sc, &stsch );
    stsch.opcode = ISCSI_OP_SCSI_CMD;
    stsch.itt = htonl(task->itt);
    stsch.cmdsn = htonl(session->CmdSn);
    stsch.expstatsn = htonl(session->ExpStatSn);

    /* set the F-bit unless unsolicited Data-out PDUs will be sent */
    if (stsch.flags.write_data && !session->initialR2T)
        stsch.flags.final = 0;
    else
        stsch.flags.final = 1;

    /* single level LUN format puts LUN in byte 1, 0 everywhere else */
    stsch.lun[1] = sc->lun;

    memcpy( stsch.scb, sc->cmnd, sc->cmd_len );

    iov.iov_base = &stsch;
    iov.iov_len = sizeof(stsch);
    memset( &msg, 0, sizeof(struct msghdr) );
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;

    ISCSI_TRACE( ISCSI_TRACE_TxCmd, sc->lun, sc->cmnd[0], sc->target,
                 (unsigned int)task, sc->bufflen );

    rc = iscsi_sendmsg( session, &msg, wlen );
    if ( rc != wlen ) {
        DEBUG_ERR2("iSCSI: xmit_task sendmsg %d failed, error %d\n",wlen,rc);
        /* close the connection */
        if (session->rx_pid)
            kill_proc(session->rx_pid, SIGPIPE, 1);
        return 0;
    }

    session->CmdSn++;

    return 1;
}


static void iscsi_xmit_queued_cmnds( iscsi_session_t *session)
{
    Scsi_Cmnd *sc;
    iscsi_task_t *task;
    cpu_flags_t cpu_flags;

    if (!session) {
        DEBUG_ERR0("iSCSI: can't xmit queued commands, no session\n");
        return;
    }

    if (test_bit(SESSION_RESETTING, &session->control_bits)) {
        DEBUG_QUEUE2("iSCSI: resetting session %p to %s, can't start tasks now\n",
                    session, session->TargetName);
        return;
    }

    DEBUG_QUEUE1("iSCSI: xmit_queued_cmnds current %p\n", current);

    spin_lock_irqsave(&session->scsi_cmnd_lock, cpu_flags);
    while (!signal_pending(current) && 
           sna_lte(session->CmdSn, session->MaxCmdSn) && 
           !test_bit(SESSION_RESETTING, &session->control_bits)) {

        /* try to grab one SCSI command off the command queue */
        if ((sc = session->scsi_cmnd_head) == NULL) {
            spin_unlock_irqrestore(&session->scsi_cmnd_lock, cpu_flags);
            return;
        }

        /* allocate a task and add it to the session before removing
         * the Scsi_Cmnd from the queue, to ensure the error handlers can
         * always find either the command or the task.
         */
        task = alloc_task(session);
        if (!task) {
            DEBUG_ERR5("iSCSI: couldn't allocate task for command %p to (%u %u %u %u)!\n",
                       sc, sc->host->host_no, sc->channel, sc->target, sc->lun);
            spin_unlock_irqrestore(&session->scsi_cmnd_lock, cpu_flags);
            /* to prevent a stall of the driver, other code must ensure that the 
             * TX thread is re-awoken when a task becomes free.
             */
            return;
        }
        task->scsi_cmnd = sc;
        atomic_inc(&task->refcount);
        add_task_to_session(session, task); /* adds to rx_tasks for us */

        /* remove Cmnd from queue */
        session->scsi_cmnd_head = (Scsi_Cmnd *)sc->host_scribble;
        /* host_scribble is now used for a task point rather than a queue pointer */
        sc->host_scribble = (unsigned char *)task;
        spin_unlock_irqrestore(&session->scsi_cmnd_lock, cpu_flags);

        DEBUG_FLOW4("iSCSI: sending itt %u on session %p as CmdSN %u, MaxCmdSn %u\n",
                     task->itt, session, session->CmdSn, session->MaxCmdSn);

        if (iscsi_xmit_task(task)) {
            if (task->flags & ISCSI_CMD_WRITE) {
                if (session->initialR2T == 0) {
                    /* send unsolicited data PDUs */
                    uint32_t length;

                    if (session->firstBurstSize)
                        task->data_length = length = MIN(session->firstBurstSize, sc->bufflen);
                    else
                        task->data_length = length = sc->bufflen;
 
                    iscsi_xmit_data(task, RSVD_TASK_TAG, 0U, length);
                } 
            }
            DEBUG_FLOW4("iSCSI: sent itt %u, task %p, cmnd %p, refcount %d\n", task->itt, task, sc, atomic_read(&task->refcount));
            atomic_dec(&task->refcount);
            DEBUG_FLOW3("iSCSI: after sending itt %u, task %p now has refcount %d\n", task->itt, task, atomic_read(&task->refcount));
        }
        else {
            DEBUG_FLOW3("iSCSI: failed to send itt %u, task %p, cmnd %p, returning to queue\n", 
                         task->itt, task, sc);

            /* get rid of the task */
            remove_task_from_session(session, task);
            atomic_dec(&task->refcount);
            free_task(task);

            /* put it back on the head of the Scsi_Cmnd queue, we couldn't send it */
            spin_lock_irqsave(&session->scsi_cmnd_lock, cpu_flags);
            if ( session->scsi_cmnd_head ) {
                sc->host_scribble = (unsigned char *)session->scsi_cmnd_head;
                session->scsi_cmnd_head = sc;
            }
            else {
                sc->host_scribble = NULL;
                session->scsi_cmnd_head = session->scsi_cmnd_tail = sc;
            }
            spin_unlock_irqrestore(&session->scsi_cmnd_lock, cpu_flags);

            /* don't try to send any more right now */
            return;
        }

        /* reacquire lock before checking loop condition */
        spin_lock_irqsave(&session->scsi_cmnd_lock, cpu_flags);
    }

    spin_unlock_irqrestore(&session->scsi_cmnd_lock, cpu_flags);
}

static void iscsi_xmit_r2t_data(iscsi_session_t *session)
{
    iscsi_task_t *task;
    uint32_t itt;
    uint32_t ttt;
    uint32_t offset;
    uint32_t length;

    spin_lock(&session->task_lock);
    while (!signal_pending(current) && (task = pop_task(&session->tx_tasks))) {
        itt = task->itt;
        /* save the values that get set when we receive an R2T before putting task in rx_tasks */
        ttt = task->ttt;
        offset = task->data_offset;
        length = task->data_length;
        atomic_inc(&task->refcount);
        add_task(&session->rx_tasks, task);
        spin_unlock(&session->task_lock);
        
        /* if we can't send any unsolicited data, try to start
         * commands as soon as possible, so that we can
         * overlap the R2T latency with the time it takes to
         * send data for commands already issued.  Command
         * PDUs are small, so this increases throughput
         * without significantly increasing the completion
         * time of commands already issued.
         */
        if (session->initialR2T && !session->immediateData) {
            DEBUG_FLOW1("iSCSI: checking for new commands before sending R2T data to %s\n", 
                        session->TargetName);
            iscsi_xmit_queued_cmnds(session);
        }
        
        /* send the requested data */
        DEBUG_FLOW4("iSCSI: sending R2T data for itt %u, ttt %u, task %p to %s\n", 
                    itt, ntohl(ttt), task, session->TargetName);
        iscsi_xmit_data(task, ttt, offset, length);

        atomic_dec(&task->refcount);
        
        /* relock before checking loop condition */
        spin_lock(&session->task_lock);
    }
    spin_unlock(&session->task_lock);
}


/* send replies to a NopIns that requested them */
static void
iscsi_xmit_nop( iscsi_session_t *session)
{
    struct IscsiNopOutHdr stnoh;
    struct msghdr msg;
    struct iovec iov;
    int rc;

    /* FIXME: we need to queue up the ttt, LUN and Nop data, so that
     * we can reply with them.  For now, send something, and hope
     * it satisfies the target.
     */
    memset( &stnoh, 0, sizeof(stnoh) );
    stnoh.opcode = ISCSI_OP_NOOP_OUT;
    stnoh.opcode |= ISCSI_OP_IMMEDIATE;
    stnoh.itt  = RSVD_TASK_TAG;
    stnoh.ttt  = RSVD_TASK_TAG; /* FIXME: ought to be the NopIn ttt */
    memset(stnoh.lun, 0, sizeof(stnoh.lun)); /* FIXME out to be the NopIn LUN */
    /* FIXME: we need to reply with the data in the NopIn, if there is any */
    stnoh.cmdsn  = htonl(session->CmdSn); /* don't increment after immediate cmds */
    stnoh.expstatsn = htonl(session->ExpStatSn);

    iov.iov_base = &stnoh;
    iov.iov_len = sizeof(stnoh);
    memset( &msg, 0, sizeof(msg) );
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;
    
    rc = iscsi_sendmsg( session, &msg, sizeof(stnoh) );
    if ( rc != sizeof(stnoh) ) {
        DEBUG_ERR1("iSCSI: xmit_nop error %d\n",rc);
        /* close the connection */
        if (session->rx_pid)
            kill_proc(session->rx_pid, SIGPIPE, 1);
    }
    
    ISCSI_TRACE( ISCSI_TRACE_TxNop, 0, 0, 0, 0, 0 );
}

/* send replies for NopIns that requested them */
static void
iscsi_xmit_nop_replys(iscsi_session_t *session)
{
    /* FIXME: this should process a queue of nop requests from the target */
    iscsi_xmit_nop(session);
}

static int
iscsi_xmit_abort(iscsi_task_t *task)
{
    struct msghdr msg;
    struct iovec iov;
    int rc;
    iscsi_session_t *session;
    struct IscsiScsiTaskMgtHdr ststmh;
    Scsi_Cmnd *sc = task->scsi_cmnd;

    session = task->session;
    if ( ! session ) {
        printk("iSCSI: no session for task %p, command %p, can't abort\n", task, sc);
        return 0;
    }

    memset( &ststmh, 0, sizeof(ststmh) );
    ststmh.opcode = ISCSI_OP_SCSI_TASK_MGT_MSG;
    /* Flag it as an Immediate CMD */
    ststmh.opcode |= ISCSI_OP_IMMEDIATE;

    /* allocate an itt.  The reply will only have the rtt
     * if the task was not found, so we need to be able to find
     * the task being aborted based on the abort's itt.
     */
    task->mgmt_itt = session->itt;
    ststmh.itt = htonl(task->mgmt_itt);
    session->itt++;
    /* iSCSI reserves 0xFFFFFFFF, this driver reserves 0 */
    if (session->itt == RSVD_TASK_TAG)
        session->itt = 1;
    ststmh.rtt = htonl(task->itt);

    ststmh.lun[1] = sc->lun;
    ststmh.function = ISCSI_TM_FUNC_ABORT_TASK;
    ststmh.cmdsn = htonl(session->CmdSn);   /* CmdSN not incremented after imm cmd */
    ststmh.expstatsn = htonl(session->ExpStatSn);

    iov.iov_base = &ststmh;
    iov.iov_len = sizeof(ststmh);
    memset( &msg, 0, sizeof(msg) );
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;

    ISCSI_TRACE( ISCSI_TRACE_TxAbort, sc->lun, sc->cmnd[0], sc->target, (unsigned int)task, 0);

    printk("iSCSI: sending abort itt %u for task %p, rtt %u\n",
           ntohl(ststmh.itt), task, task->itt);

    /* indicate that we're done with the task */
    atomic_dec(&task->refcount);
    task = NULL;
    sc = NULL;

    /* send the abort */
    rc = iscsi_sendmsg( session, &msg, sizeof(ststmh) );
    if ( rc != sizeof(ststmh) ) {
        DEBUG_ERR3("iSCSI: xmit_abort failed, itt %u, rtt %u, rc %d\n", 
                   ntohl(ststmh.itt), ntohl(ststmh.rtt), rc);
        /* close the connection */
        if (session->rx_pid)
            kill_proc(session->rx_pid, SIGPIPE, 1);
        return 0;
    }

    return 1;
}

/* send aborts for every task that needs one */
static void iscsi_xmit_aborts(iscsi_session_t *session) 
{
    iscsi_task_t *task;

    spin_lock(&session->task_lock);
    while (!signal_pending(current) && (task = pop_task(&session->tx_abort_tasks))) {
        atomic_inc(&task->refcount);
        add_task(&session->rx_abort_tasks, task);
        spin_unlock(&session->task_lock);

        iscsi_xmit_abort(task);
        /* xmit_abort decrements the task refcount when it's done with it */

        /* relock before checking loop condition */
        spin_lock(&session->task_lock);
    }
    spin_unlock(&session->task_lock);
}

static int
iscsi_xmit_reset(iscsi_session_t *session, uint8_t reset_type, iscsi_task_t *task)
{
    struct msghdr msg;
    struct iovec iov;
    int rc;
    struct IscsiScsiTaskMgtHdr ststmh;
    Scsi_Cmnd *sc;

    memset( &ststmh, 0, sizeof(ststmh) );
    ststmh.opcode = ISCSI_OP_SCSI_TASK_MGT_MSG;
    /* Flag it as an Immediate CMD */
    ststmh.opcode |= ISCSI_OP_IMMEDIATE;

    /* record the itt */
    if (task)
        task->mgmt_itt = session->itt;
    else if (reset_type == ISCSI_TM_FUNC_TARGET_WARM_RESET)
        session->warm_reset_itt = session->itt;
    else if (reset_type == ISCSI_TM_FUNC_TARGET_COLD_RESET)
        session->cold_reset_itt = session->itt;

    ststmh.itt = htonl(session->itt); 
    session->itt++;
    /* iSCSI reserves 0xFFFFFFFF, this driver reserves 0 */
    if (session->itt == RSVD_TASK_TAG)
        session->itt = 1;

    ststmh.function = reset_type;
    ststmh.cmdsn = htonl(session->CmdSn);   /* CmdSN not incremented after imm cmd */
    ststmh.expstatsn = htonl(session->ExpStatSn);

    switch (reset_type) {
        case ISCSI_TM_FUNC_LOGICAL_UNIT_RESET:
            /* need a LUN for this */
            if (task && task->scsi_cmnd) {
                sc = task->scsi_cmnd;
                ststmh.lun[1] = sc->lun;
                printk("iSCSI: sending logical unit reset to (%u %u %u %u), itt %u\n", 
                       sc->host->host_no, sc->channel, sc->target, sc->lun, ntohl(ststmh.itt));
            }
            else {
                printk("iSCSI: failed to send logical unit reset, no SCSI command\n");
                return 0;
            }
            break;
        case ISCSI_TM_FUNC_TARGET_WARM_RESET:
            printk("iSCSI: sending target warm reset for session %p to %s, itt %u\n", 
                   session, session->TargetName, ntohl(ststmh.itt));
            break;
        case ISCSI_TM_FUNC_TARGET_COLD_RESET:
            printk("iSCSI: sending target cold reset for session %p to %s, itt %u\n", 
                   session, session->TargetName, ntohl(ststmh.itt));
            break;
        default:
            printk("iSCSI: unknown reset type %u for session %p to %s\n", 
                   reset_type, session, session->TargetName);
            return 0;
            break;
    }

    iov.iov_base = &ststmh;
    iov.iov_len = sizeof(ststmh);
    memset( &msg, 0, sizeof(msg) );
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;

    /* send the reset */
    rc = iscsi_sendmsg( session, &msg, sizeof(ststmh) );
    if ( rc != sizeof(ststmh) ) {
        DEBUG_ERR1("iSCSI: xmit_reset failed, rc %d\n", rc);
        /* close the connection */
        if (session->rx_pid)
            kill_proc(session->rx_pid, SIGPIPE, 1);
        return 0;
    }
    
    return 1;
}

static void iscsi_xmit_lun_resets(iscsi_session_t *session)
{
    iscsi_task_t *task;

    spin_lock(&session->task_lock);
    while (!signal_pending(current) && (task = pop_task(&session->tx_lun_reset_tasks))) {
        atomic_inc(&task->refcount);
        add_task(&session->rx_tasks, task);
        spin_unlock(&session->task_lock);

        iscsi_xmit_reset(session, ISCSI_TM_FUNC_LOGICAL_UNIT_RESET, task);
        atomic_dec(&task->refcount);

        /* relock before checking loop condition */
        spin_lock(&session->task_lock);
    }
    spin_unlock(&session->task_lock);
}


static void
iscsi_xmit_ping( iscsi_session_t *session )
{
    struct IscsiNopOutHdr stph;
    struct msghdr msg;
    struct iovec iov;
    int rc;
    
    memset( &stph, 0, sizeof(stph) );
    stph.opcode = ISCSI_OP_NOOP_OUT;
    stph.opcode |= ISCSI_OP_IMMEDIATE;
    stph.poll = 1;  /* draft 8 always wants this bit set now */
    stph.itt  = htonl(session->itt); /* draft 8 reply request */
    session->itt++;
    /* iSCSI reserves 0xFFFFFFFF, this driver reserves 0 */
    if (session->itt == RSVD_TASK_TAG)
        session->itt = 1;
    stph.ttt  = RSVD_TASK_TAG;
    stph.cmdsn = htonl(session->CmdSn);
    stph.expstatsn = htonl(session->ExpStatSn);

    iov.iov_base = &stph;
    iov.iov_len = sizeof(stph);
    memset( &msg, 0, sizeof(msg) );
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;
    
    rc = iscsi_sendmsg( session, &msg, sizeof(stph) );
    if ( rc != sizeof(stph) ) {
        DEBUG_ERR1("iSCSI: XmitPing error %d\n",rc);
        /* close the connection */
        if (session->rx_pid)
            kill_proc(session->rx_pid, SIGPIPE, 1);
    }
    
    ISCSI_TRACE( ISCSI_TRACE_TxPing, 0, 0, 0, 0, 0 );
}


/* the writer thread */
static int
iscsi_tx_thread( void *vtaskp )
{
    iscsi_session_t *session;

    if ( ! vtaskp ) {
        DEBUG_ERR0("iSCSI: tx thread task parameter NULL\n");
        return 0;
    }

    session = (iscsi_session_t *)vtaskp;
    
    /* check for any bits currently set on the first iteration */
    set_bit(TX_WAKE, &session->control_bits);

    DEBUG_INIT2("iSCSI: tx thread to %s, current %p\n", session->TargetName, current);

    while (!signal_pending(current)) {

        DEBUG_FLOW1("iSCSI: tx thread %s waiting\n", session->TargetName);
        wait_event_interruptible(session->tx_wait_q, test_and_clear_bit(TX_WAKE, &session->control_bits));
        DEBUG_FLOW1("iSCSI: tx thread %s is awake\n", session->TargetName);

        if ( signal_pending(current) ) {
            DEBUG_FLOW1("iSCSI: tx thread %s received signal\n", session->TargetName);
            break;
        }

        if (session->tx_pid == 0)
            session->tx_pid = current->pid;

        /* See if we should send a ping (Nop with reply requested) */
        if (test_and_clear_bit(TX_PING, &session->control_bits)) {
            DEBUG_FLOW1("iSCSI: sending Nop/poll to %s\n", session->TargetName);
            iscsi_xmit_ping( session );
        }

        /* See if we should send one or more Nops (replies requested by the target) */
        if (test_and_clear_bit(TX_NOP_REPLY, &session->control_bits)) {
            DEBUG_FLOW1("iSCSI: sending Nop replies to %s\n", session->TargetName);
            iscsi_xmit_nop_replys(session);
        }

        /* See if we should abort any tasks */
        if (test_and_clear_bit(TX_ABORT, &session->control_bits)) {
            DEBUG_FLOW1("iSCSI: sending aborts to %s\n", session->TargetName);
            iscsi_xmit_aborts(session);
        }

        /* See if we should reset any LUs */
        if (test_and_clear_bit(TX_LUN_RESET, &session->control_bits)) {
            DEBUG_FLOW1("iSCSI: sending logical unit resets to %s\n", session->TargetName);
            iscsi_xmit_lun_resets(session);
        }

        /* See if we should warm reset the target */
        if (test_and_clear_bit(TX_WARM_TARGET_RESET, &session->control_bits)) {
            DEBUG_FLOW1("iSCSI: sending target warm reset to %s\n", session->TargetName);
            iscsi_xmit_reset(session, ISCSI_TM_FUNC_TARGET_WARM_RESET, NULL);
        }

        /* See if we should cold reset the target */
        if (test_and_clear_bit(TX_COLD_TARGET_RESET, &session->control_bits)) {
            DEBUG_FLOW1("iSCSI: sending target cold reset to %s\n", session->TargetName);
            iscsi_xmit_reset(session, ISCSI_TM_FUNC_TARGET_COLD_RESET, NULL);
        }

        /* See if we need to send more data after receiving an R2T */
        if (test_and_clear_bit(TX_R2T_DATA, &session->control_bits)) {
            /* NOTE: this may call iscsi_xmit_queued_cmnds under some conditions */
            iscsi_xmit_r2t_data(session);
        }

        /* New SCSI command received, or MaxCmdSN incremented, or task freed */
        if (test_and_clear_bit(TX_SCSI_COMMAND, &session->control_bits)) {
            /* if possible, issue new commands */
            iscsi_xmit_queued_cmnds(session);
        }
    }

    DEBUG_INIT1("iSCSI: tx thread to %s shutting down\n", session->TargetName);

    /* kill the reader thread, if there still is one */
    if ( session->rx_pid > 1) {
        DEBUG_INIT2("iSCSI: tx thread %d killing rx thread %d\n", session->tx_pid, session->rx_pid);
        kill_proc(session->rx_pid, SIGPIPE, 1);
    }

    set_current_state(TASK_RUNNING);
    DEBUG_INIT1("iSCSI: tx thread for %s exiting\n", session->TargetName);
    
    return 0;
}

/* 
 * complete a task in the session's completing queue, and return a pointer to it,
 * or NULL if the task could not be completed.
 */
static iscsi_task_t *complete_task(iscsi_session_t *session, uint32_t itt)
{
    iscsi_task_t *task;
    unsigned long last_log = 0;
    unsigned long interval = (HZ / 10) ? (HZ / 10) : 10;
    int refcount;
    DECLARE_CPU_FLAGS;

    while (!signal_pending(current)) {
        DEBUG_QUEUE1("iSCSI: attempting to complete itt %u\n", itt);

        LOCK_IO_REQUEST_LOCK;
        DEBUG_FLOW1("iSCSI: acquired io_request_lock for completing itt %u\n", itt);

        spin_lock(&session->task_lock);
        DEBUG_FLOW1("iSCSI: acquired task_lock for completing itt %u\n", itt);

        if ((task = remove_task(&session->completing_tasks, itt))) {
            Scsi_Cmnd *sc = task->scsi_cmnd;

            if (sc == NULL) {
                DEBUG_QUEUE2("iSCSI: can't complete itt %u, task %p, command already aborted\n",
                            itt, task);
                spin_unlock(&session->task_lock);
                UNLOCK_IO_REQUEST_LOCK;
                return NULL;
            }

            if (task->mgmt_itt != RSVD_TASK_TAG) {
                DEBUG_QUEUE4("iSCSI: can't complete itt %u, task %p, cmnd %p, aborting with mgmt_itt %u\n",
                             itt, task, sc, task->mgmt_itt);
                spin_unlock(&session->task_lock);
                UNLOCK_IO_REQUEST_LOCK;
                return NULL;
            }

            /* it's possible the tx thread is using the task right now.
             * since the task is now in the completing collection, it's refcount can't
             * increase.  Wait for the refcount to reach zero before completing.
             */
            if ((refcount = atomic_read(&task->refcount)) == 0) {
                /* this is the expected case */
#ifdef DEBUG_QUEUE
                printk("iSCSI: completing itt %u, task %p, command %p, (%u %u %u %u), Cmd 0x%x, result 0x%x\n", 
                       itt, task, sc, 
                       sc->host->host_no, sc->channel, sc->target, sc->lun, 
                       sc->cmnd[0], sc->result);
#endif
                if (sc->scsi_done)
                    sc->scsi_done(sc);

                spin_unlock(&session->task_lock);
                UNLOCK_IO_REQUEST_LOCK;

                ISCSI_TRACE( ISCSI_TRACE_CmdDone, task->scsi_cmnd->lun, task->scsi_cmnd->cmnd[0], 
                             task->scsi_cmnd->target, (unsigned int)task, 0 );

                return task;
            }
            else {
                /* task is still in use, can't complete it yet.  Since
                 * this only happens when a command is aborted by the
                 * target unexpectedly, this error case can be slow.
                 * Just keep polling for the refcount to hit zero.  If
                 * the tx thread is blocked while using a task, the
                 * timer thread will eventually send a signal to both
                 * the rx thread and tx thread, so this loop will
                 * terminate one way or another.  
                 */
                if ((last_log == 0) || jiffy_lt(last_log + HZ, jiffies)) {
                    DEBUG_QUEUE4("iSCSI: waiting to complete itt %u, task %p, cmnd %p, refcount %d\n", itt, task, sc, refcount);
                }

                push_task(&session->completing_tasks, task);
                spin_unlock(&session->task_lock);
                UNLOCK_IO_REQUEST_LOCK;

                set_current_state(TASK_INTERRUPTIBLE);
                schedule_timeout(interval);
            }
        }
        else {
            /* an abort removed it from the completing collection. */
            DEBUG_QUEUE1("iSCSI: can't complete itt %u, task no longer completing\n", itt);
            spin_unlock(&session->task_lock);
            UNLOCK_IO_REQUEST_LOCK;
            return NULL;
        }
    }

    return NULL;
}


static void iscsi_recv_nop(iscsi_session_t *session, struct IscsiNopInHdr *stnih)
{
    ISCSI_TRACE( ISCSI_TRACE_RxPing, 0, 0, 0, 0, 0 );
    DEBUG_FLOW2("iSCSI: recv_nop for session %p from %s\n", session, session->TargetName);

    session->ExpStatSn = ntohl(stnih->statsn) + 1;
    updateSN(session, ntohl(stnih->expcmdsn), ntohl(stnih->maxcmdsn));

    /* check the ttt to decide whether to reply with a Nop-out */
    if (stnih->ttt != RSVD_TASK_TAG) {
        /* FIXME: need to make the tx thread send a Nop reply, with
         * whatever ttt, lun, and data was in this NopIn.  Since the
         * tx thread may be busy, we may need to queue up the ping
         * (and it's data) til later.  For now, just send back
         * something, and hope the target doesn't care about the
         * things we're not sending back.  
         */
        printk("iSCSI: received Nop, ttt %u, sending reply\n", ntohl(stnih->ttt));
        wake_tx_thread(TX_NOP_REPLY, session);
    }
}

static void
iscsi_recv_cmd(iscsi_session_t *session, struct IscsiScsiRspHdr *stsrh, unsigned char *xbuf )
{
    iscsi_task_t *task;
    Scsi_Cmnd *sc;
    unsigned char *respbuf;
    int senselen;
    uint32_t itt = ntohl(stsrh->itt);

    updateSN(session, ntohl(stsrh->expcmdsn), ntohl(stsrh->maxcmdsn));

    /* find the task for the itt we received */
    spin_lock(&session->task_lock);
    if ((task = remove_task(&session->rx_tasks, itt))) {
        /* task was waiting for this command response */
        DEBUG_QUEUE3("iSCSI: recv_cmd - rx_tasks has itt %u, task %p, refcount %d\n", itt, task, atomic_read(&task->refcount));
        atomic_inc(&task->refcount);
        add_task(&session->completing_tasks, task);
    }
    else if ((task = remove_task(&session->tx_tasks, itt))) {
        /* target aborted the command for some reason, even
         * though we're trying to send it more data.  
         */
        DEBUG_QUEUE3("iSCSI: recv_cmd - tx_tasks has itt %u, task %p, refcount %d\n", itt, task, atomic_read(&task->refcount));
        atomic_inc(&task->refcount);
        add_task(&session->completing_tasks, task);
    }
    else if ((task = find_task(&session->rx_abort_tasks, itt))) {
        /* just leave it waiting for an abort response */
        spin_unlock(&session->task_lock);
        DEBUG_INIT3("iSCSI: recv_cmd - ignoring cmd response for itt %u, task %p, command %p, waiting for abort response\n", 
                    itt, task, task->scsi_cmnd);
        return;
    }
    else if ((task = find_task(&session->tx_abort_tasks, itt))) {
        /* don't bother sending an abort, since the command has already completed.
         * we'd just get an abort rejected response.
         */
        spin_unlock(&session->task_lock);
        DEBUG_INIT3("iSCSI: recv_cmd ignoring cmd response for itt %u, task %p, command %p, abort queued\n",
                    itt, task, task->scsi_cmnd);
        return;
    }
    spin_unlock(&session->task_lock);

    if (!task) {
        /* odd, there should be a task somewhere. */
        DEBUG_ERR1("iSCSI: recv_cmd - response for itt %u, but no such task\n", itt);
        return;
    }

    sc = task->scsi_cmnd;

    DEBUG_FLOW6("iSCSI: recv_cmd - itt %u, task %p, cmnd %p, Cmd 0x%x, cmd_len %d, rsp dlength %d\n",
                itt, task, sc, sc->cmnd[0], sc->cmd_len, ntoh24(stsrh->dlength));

    if (stsrh->response) {
        /*
         * FIXME: This may contain sense data to indicate certain errors as
         * described in sec 2.4.3 of the iSCSI protocol
         */                                                                            
        /*
         * If this is not giving the right status, remove this
         * check and let the cmd_status check handle it.
         */
        ISCSI_TRACE( ISCSI_TRACE_RxCmdStatus, sc->lun, sc->cmnd[0], sc->target,
                     (unsigned int)task, stsrh->cmd_status );
        DEBUG_INIT3("iSCSI: recv_cmd %p, response 0x%x, cmd_status 0x%x\n", sc, stsrh->response, stsrh->cmd_status);
        sc->result = HOST_BYTE(DID_NO_CONNECT) | STATUS_BYTE(stsrh->cmd_status);  
        /* FIXME: use DID_SOFT_ERROR? do we want a retry? */
    } 
    else {
        sc->result = HOST_BYTE(DID_OK) | STATUS_BYTE(stsrh->cmd_status);
        if ( stsrh->cmd_status ) {
            senselen = ntoh24(stsrh->dlength);
            ISCSI_TRACE( ISCSI_TRACE_RxCmdCmdStatus, sc->lun, sc->cmnd[0],
                         sc->target, (unsigned int)task, stsrh->cmd_status );
            if ( senselen) {
                if (senselen >= 26) {
                    printk("iSCSI: recv_cmd %p, Cmd 0x%x, status 0x%x, senselen %d, task %p to (%u %u %u %u), %s\n"
                           "       Sense %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x " 
                           "%02x%02x%02x%02x %02x%02x%02x%02x %02x%02x\n",
                           sc, sc->cmnd[0], stsrh->cmd_status, senselen, 
                           task, sc->host->host_no, sc->channel, sc->target, sc->lun, session->TargetName,
                           xbuf[0],xbuf[1],xbuf[2],xbuf[3],
                           xbuf[4],xbuf[5],xbuf[6],xbuf[7],
                           xbuf[8],xbuf[9],xbuf[10],xbuf[11],
                           xbuf[12],xbuf[13],xbuf[14],xbuf[15],
                           xbuf[16],xbuf[17],xbuf[18],xbuf[19],
                           xbuf[20],xbuf[21],xbuf[22],xbuf[23],
                           xbuf[24], xbuf[25]);
                }
                else if ( senselen >= 18) {
                    printk("iSCSI: recv_cmd %p, Cmd 0x%x, status 0x%x, senselen %d, task %p to (%u %u %u %u), %s\n"
                           "       Sense %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x\n",
                           sc, sc->cmnd[0], stsrh->cmd_status, senselen, 
                           task, sc->host->host_no, sc->channel, sc->target, sc->lun, session->TargetName,
                           xbuf[0],xbuf[1],xbuf[2],xbuf[3],
                           xbuf[4],xbuf[5],xbuf[6],xbuf[7],
                           xbuf[8],xbuf[9],xbuf[10],xbuf[11],
                           xbuf[12],xbuf[13],xbuf[14],xbuf[15],
                           xbuf[16],xbuf[17]);
                } 
                if ( senselen > sizeof(sc->sense_buffer) ) {
                    senselen = sizeof(sc->sense_buffer);
                }
                memcpy( sc->sense_buffer, xbuf, senselen );
            }
        }
    }

    if ( (task->flags & ISCSI_CMD_READ) &&
         (task->rxdata != sc->bufflen) ) {

        if (stsrh->flags.overflow || stsrh->flags.underflow ) {
            DEBUG_INIT5("iSCSI: Cmd 0x%x, %s, received %d, residual %d, bufflen %d\n",
                        sc->cmnd[0], stsrh->flags.overflow ? "overflow" : "underflow",
                        task->rxdata, ntohl(stsrh->residual_count), sc->bufflen);
            if ( stsrh->flags.underflow ) {
                ISCSI_TRACE( ISCSI_TRACE_RxCmdFlow, sc->lun, sc->cmnd[0],
                             sc->target, (unsigned int)task, ntohl(stsrh->residual_count) );
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,19)
                sc->resid = ntohl(stsrh->residual_count);
#else
                if ( task->rxdata < sc->underflow ) {
                    sc->result = HOST_BYTE(DID_ERROR) | STATUS_BYTE(stsrh->cmd_status);
                }
#endif
            } else {
                ISCSI_TRACE( ISCSI_TRACE_RxCmdFlow, sc->lun, sc->cmnd[0],
                             sc->target, (unsigned int)task, 0 );
            }
        } 
        else {
            /* All the data did not arrive */
            DEBUG_INIT4("iSCSI: task %p, cmnd %p received only %d of %d\n",
                        task, sc, task->rxdata, sc->bufflen);
            sc->result = HOST_BYTE(DID_ERROR) | STATUS_BYTE(stsrh->cmd_status);
        }
    }

    ISCSI_TRACE( ISCSI_TRACE_RxCmd, sc->lun, sc->cmnd[0], sc->target,(unsigned int)task, task->rxdata);

    if ( sc->cmnd[0] == TEST_UNIT_READY ) {
        if ( stsrh->cmd_status == 0 ) {
            /* when a LUN becomes ready */
            if ( test_and_set_bit(sc->lun, session->lun_bitmap)) {
                DEBUG_FLOW5("iSCSI: unit still ready, (%u %u %u %u), TargetName %s\n",
                            sc->host->host_no, sc->channel, sc->target, sc->lun, session->TargetName);
            }
            else {
                printk("iSCSI: unit ready, (%u %u %u %u), TargetName %s\n",
                       sc->host->host_no, sc->channel, sc->target, sc->lun, session->TargetName);
            }
        }
        else {
            if (test_and_clear_bit(sc->lun, session->lun_bitmap)) {
                /* was ready, but isn't anymore */
                printk("iSCSI: unit no longer ready, (%u %u %u %u), TargetName %s\n",
                       sc->host->host_no, sc->channel, sc->target, sc->lun, session->TargetName);
            }
            else {
                DEBUG_FLOW5("iSCSI: unit not ready, (%u %u %u %u), TargetName %s\n",
                            sc->host->host_no, sc->channel, sc->target, sc->lun, session->TargetName);
            }
        }

        respbuf = sc->buffer;
        memset( respbuf, 0, sc->bufflen );
    }

#if TEST_ABORTS
    if (session->rx_pid && sc &&
        (sc->cmnd[0] != TEST_UNIT_READY) &&
        (stsrh->cmd_status == 0) &&
        (task->cmdsn % ABORT_FREQUENCY) == 0) {
        /* don't complete this command, so that we can test the error handling
         * code.
         */
        spin_lock(&session->task_lock);
        if ((task = remove_task(&session->completing_tasks, itt))) {
            add_task(&session->rx_tasks, task);
            printk("iSCSI: ignoring completion of itt %u, task %p, sc %p, Cmd 0x%x to (%u %u %u %u)\n", 
                   itt, task, sc, sc->cmnd[0], sc->host->host_no, sc->channel, sc->target, sc->lun);
        }
        spin_unlock(&session->task_lock);
        atomic_dec(&task->refcount);
        return;
    }
#endif

    /* we're done using the command and task */
    atomic_dec(&task->refcount);
    task = NULL;
    sc = NULL;

    /* now that we're done with it, try to complete it.  This may fail
     * if an abort for this command comes in while we're not holding
     * certain locks.  That's ok.  
     */
    DEBUG_FLOW1("iSCSI: recv_cmd attempting to complete itt %u\n", itt);
    if ((task = complete_task(session, itt))) {
        remove_task_from_session(session, task);
        free_task(task);
    }
}

static void
iscsi_recv_r2t(iscsi_session_t *session, struct IscsiRttHdr *strh )
{
    iscsi_task_t *task = NULL;
    uint32_t itt = ntohl(strh->itt);
    uint32_t offset = 0, length = 0;
    uint32_t ttt = RSVD_TASK_TAG;

    updateSN(session, ntohl(strh->expcmdsn), ntohl(strh->maxcmdsn));

    spin_lock(&session->task_lock);
    if ((task = remove_task(&session->rx_tasks, itt))) {
        task->ttt = ttt = strh->ttt;
        task->data_length = length = ntohl(strh->data_length);
        task->data_offset = offset = ntohl(strh->data_offset);
        add_task(&session->tx_tasks, task);
        ISCSI_TRACE( ISCSI_TRACE_RttData, sc->lun, sc->cmnd[0], sc->target, (unsigned int)task, ttt);
    }
    spin_unlock(&session->task_lock);

    if (task) {
        DEBUG_FLOW4("iSCSI: R2T for task %p itt %u, %u bytes @ offset %u\n",
                    task, ntohl(strh->itt), ntohl(strh->data_length), ntohl(strh->data_offset));
        
        /* wake up the Tx thread for this connection */
        wake_tx_thread(TX_R2T_DATA, session);
    }
    else {
        /* the task wasn't waiting for an R2T, aborting or already received an R2T */
        DEBUG_ERR3("iSCSI: ignoring R2T for itt %u, %u bytes @ offset %u\n",
                   ntohl(strh->itt), ntohl(strh->data_length), ntohl(strh->data_offset));
    }
}


static void
iscsi_recv_data( iscsi_session_t *session, struct IscsiDataRspHdr *stdrh )
{
    iscsi_task_t *task = NULL;
    Scsi_Cmnd *sc = NULL;
    struct scatterlist *sglist = NULL;
    int i, dlength, rc;
    uint32_t offset;
    unsigned int segNum = 0, iovn, pad, relative_offset;
    struct msghdr msg;
    struct iovec iov;
    uint32_t itt = ntohl(stdrh->itt);
    bool found_segment = false;

    updateSN(session, ntohl(stdrh->expcmdsn), ntohl(stdrh->maxcmdsn));

    dlength = ntoh24( stdrh->dlength );
    offset = ntohl( stdrh->offset );
    /* Compute padding bytes that follow the data */
    pad = dlength % PAD_WORD_LEN;
    if (pad) {
        pad = PAD_WORD_LEN - pad;
    }

    spin_lock(&session->task_lock);
    if ((task = find_task(&session->rx_tasks, itt))) {
        /* receive the data, and leave it in rx_tasks for more data or a cmd response */
        atomic_inc(&task->refcount);
        sc = task->scsi_cmnd;
    }
    else if ((task = find_task(&session->rx_abort_tasks, itt))) {
        /* discard the data */
        DEBUG_ERR1("iSCSI: RxData - task aborting, itt %u, ignoring received data\n", ntohl(stdrh->itt));
        task = NULL;
    }
    else if ((task = find_task(&session->tx_abort_tasks, itt))) {
        /* discard the data */
        DEBUG_ERR1("iSCSI: RxData - task will abort, itt %u, ignoring received data\n", ntohl(stdrh->itt));
        task = NULL;
    }
    else {
        DEBUG_ERR1("iSCSI: RxData - no task for itt %u, ignoring received data\n", ntohl(stdrh->itt));
        task = NULL;
    }
    spin_unlock(&session->task_lock);

    /* no SCSI command, just throw away the PDU */
    if (!sc) {
        int bytes_read = 0;

        while (!signal_pending(current) && (bytes_read < dlength + pad)) {
            int num_bytes = MIN(dlength + pad - bytes_read, sizeof(session->RxBuf));
            iov.iov_base = session->RxBuf;
            iov.iov_len = sizeof(session->RxBuf);;
            memset( &msg, 0, sizeof(struct msghdr) );
            msg.msg_iov = &iov;
            msg.msg_iovlen = 1;
            rc = iscsi_recvmsg( session, &msg, num_bytes);
            if ( rc != num_bytes ) {
                DEBUG_ERR2("iSCSI: failed to recv %d data PDU bytes, rc %d\n", num_bytes, rc);
                if (session->rx_pid)
                    kill_proc(session->rx_pid, SIGPIPE, 1);
            }
            bytes_read += num_bytes;
        }
        return;
    }

    relative_offset = offset;
    
    if (sc->use_sg) {
        /* scatter-gather */

        sglist = (struct scatterlist *)sc->buffer;
        
        /* We need to find the sg segment where this data should go.
         * We look at the offset and dlength to find it (amg)
         */
        /* FIXME: make sure we don't walk off the end of the sg buffer */
        do {
            if (relative_offset < sglist->length ) {
                found_segment = true;
            } 
            else {
                relative_offset -= sglist->length;
                sglist++;
                segNum++;
            }
        } while (!found_segment);
        
        iovn = sc->use_sg - segNum;
        
        for (i = 0; i < iovn; i++) {
            if (i == 0) {
                session->RxIov[i].iov_base = sglist->address + relative_offset;
                session->RxIov[i].iov_len  = sglist->length  - relative_offset;
            } 
            else {
                session->RxIov[i].iov_base = sglist->address;
                session->RxIov[i].iov_len  = sglist->length;
            }
            sglist += 1;
        }
    } 
    else {
        /* FIXME: make sure we don't go off the end of the buffer */
        /* no scatter-gather, just read it into the buffer */
        session->RxIov[0].iov_base = sc->buffer + offset;
        session->RxIov[0].iov_len  = dlength;
        iovn = 1;
    }

    memset( &msg, 0, sizeof(struct msghdr) );
    msg.msg_iov = session->RxIov;
    msg.msg_iovlen = iovn;
    
    rc = iscsi_recvmsg( session, &msg, dlength );
    if ( rc != dlength ) {
        DEBUG_ERR2("iSCSI: failed to recv %d data PDU bytes, rc %d\n", dlength, rc);
        if (session->rx_pid)
            kill_proc(session->rx_pid, SIGPIPE, 1);
        atomic_dec(&task->refcount);
        return;
    }
    
    task->rxdata += rc;
    
    ISCSI_TRACE( ISCSI_TRACE_RxData, sc->lun, sc->cmnd[0], sc->target, (unsigned int)task, dlength);

    /* FIXME: Really ought to read this with the same recvmsg call as above.
     * a 2nd recvmsg for 1-3 bytes is almost entirely overhead, and bad for performance.
     */
    if (pad) {
        iov.iov_base = session->RxBuf;
        iov.iov_len = pad;
        memset( &msg, 0, sizeof(struct msghdr) );
        msg.msg_iov = &iov;
        msg.msg_iovlen = 1;
        rc = iscsi_recvmsg( session, &msg, pad );
        if ( rc != pad ) {
            DEBUG_ERR2("iSCSI: failed to recv %d pad bytes, rc %d\n", pad, rc);
            if (session->rx_pid)
                kill_proc(session->rx_pid, SIGPIPE, 1);
            atomic_dec(&task->refcount);
            return;
        }
    }

    if (stdrh->status_present) {
        /* we got status, meaning the command completed in a way that
         * doesn't give us any sense data, and the command must be
         * completed now, since we won't get a command response PDU.
         */
        /* FIXME: check the StatSN */
        DEBUG_FLOW4("iSCSI: Data-in with status 0x%x for itt %u, task %p, sc %p\n", 
                    stdrh->cmd_status, ntohl(stdrh->itt), task, task->scsi_cmnd);
        sc->result = HOST_BYTE(DID_OK) | STATUS_BYTE(stdrh->cmd_status);
        if ( (task->flags & ISCSI_CMD_READ) &&
             (task->rxdata != sc->bufflen) ) {

            if (stdrh->overflow || stdrh->underflow ) {
                DEBUG_INIT5("iSCSI: Cmd 0x%x, %s, received %d, residual %d, bufflen %d\n",
                            sc->cmnd[0], stdrh->overflow ? "overflow" : "underflow", 
                            task->rxdata, ntohl(stdrh->residual_count), sc->bufflen);

                if ( stdrh->underflow ) {
                    ISCSI_TRACE( ISCSI_TRACE_RxCmdFlow, sc->lun, sc->cmnd[0],
                                 sc->target, (unsigned int)task, ntohl(stdrh->residual_count) );
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,19)
                    sc->resid = ntohl(stdrh->residual_count);
#else
                    if ( task->rxdata < sc->underflow ) {
                        sc->result = HOST_BYTE(DID_ERROR);
                    }
#endif
                } 
                else {
                    ISCSI_TRACE( ISCSI_TRACE_RxCmdFlow, sc->lun, sc->cmnd[0],
                                 sc->target, (unsigned int)task, 0 );
                }
            } 
            else {
                /* All the data did not arrive */
                sc->result = HOST_BYTE(DID_ERROR) | STATUS_BYTE(stdrh->cmd_status); 
            }
        }


#if TEST_ABORTS
        if (session->rx_pid &&
            (sc->cmnd[0] != TEST_UNIT_READY) &&
            (stdrh->cmd_status == 0) &&
            (task->cmdsn % ABORT_FREQUENCY) == 0) {
            /* don't complete this command, so that we can test the error handling
             * code.
             */
            printk("iSCSI: ignoring completion of itt %u, task %p, sc %p, Cmd 0x%x to (%u %u %u %u)\n", 
                   itt, task, sc, sc->cmnd[0], sc->host->host_no, sc->channel, sc->target, sc->lun);
            atomic_dec(&task->refcount);
            return;
        }
#endif

        /* done using the command and task */
        atomic_dec(&task->refcount);

        /* try to complete the task */
        spin_lock(&session->task_lock);
        if ((task = remove_task(&session->rx_tasks, itt))) {
            add_task(&session->completing_tasks, task);
            spin_unlock(&session->task_lock);

            if ((task = complete_task(session, itt))) {
                remove_task_from_session(session, task);
                free_task(task);
            }
            else {
                DEBUG_FLOW1("iSCSI: Data-in with status for itt %u, but task couldn't be completed\n", 
                            ntohl(stdrh->itt));
            }
        }
        else {
            /* task was probably aborted */
            spin_unlock(&session->task_lock);
            DEBUG_FLOW1("iSCSI: Data-in with status for itt %u, but task aborted\n", ntohl(stdrh->itt));
        }
    }
    else {
        /* done modifying the command and task */
        atomic_dec(&task->refcount);
    }
}

static void iscsi_recv_mgmt_rsp(iscsi_session_t *session, struct IscsiScsiTaskMgtRspHdr *ststmrh )
{
    iscsi_task_t *task = NULL;
    uint32_t mgmt_itt = ntohl(ststmrh->itt);

    updateSN(session, ntohl(ststmrh->expcmdsn), ntohl(ststmrh->maxcmdsn));

    /* FIXME: we need to determine what the task mgmt function this is
     * a response to based solely on the itt.  The RTT is usually not
     * set, except for aborts of tasks that weren't found.  We don't
     * get the mgmt function back, so we need to figure out if it's an
     * abort or reset response based on the itt.  For now, rely on
     * using only aborts (which get a resonse), and cold target
     * resets, where we don't need to care about a response, since the
     * target MUST drop the connection.  
     */
    spin_lock(&session->task_lock);
    if (mgmt_itt == session->warm_reset_itt) {
        /* warm reset occured */
        session->warm_reset_itt = RSVD_TASK_TAG;
        if ( ststmrh->response ) {
            printk("iSCSI: warm reset rejected (0x%x) for itt %u, rtt %u, session %p to %s\n", 
                   ststmrh->response, ntohl(ststmrh->itt), ntohl(ststmrh->rtt), session, session->TargetName);
            clear_bit(SESSION_RESETTING, &session->control_bits);
            wake_tx_thread(TX_SCSI_COMMAND, session);
        }
        else {
            /* clear out the task collections, to make sure any command completions
             * that arrive get ignored.
             */
            session->rx_tasks.head = session->rx_tasks.tail = NULL;
            session->tx_tasks.head = session->tx_tasks.tail = NULL;
            session->rx_abort_tasks.head = session->rx_abort_tasks.tail = NULL;
            session->tx_abort_tasks.head = session->tx_abort_tasks.tail = NULL;
            session->aborted_tasks.head = session->aborted_tasks.tail = NULL;
            session->completing_tasks.head = session->completing_tasks.tail = NULL;
            session->tx_lun_reset_tasks.head = session->tx_lun_reset_tasks.tail = NULL;
            set_bit(SESSION_RESET, &session->control_bits);
            clear_bit(SESSION_RESETTING, &session->control_bits);
            wake_tx_thread(TX_SCSI_COMMAND, session);
            printk("iSCSI: warm reset success for itt %u, rtt %u, session %p to %s\n",
                   ntohl(ststmrh->itt), ntohl(ststmrh->rtt), session, session->TargetName);
        }
    }
    else if ((task = remove_mgmt_task(&session->rx_abort_tasks, mgmt_itt))) {
        /* We don't really care if the target actually aborted the
         * command, we just want it to be done so that we can free the
         * task before returning from eh_abort.
         */
        add_task(&session->aborted_tasks, task);
        if ( ststmrh->response ) {
            printk("iSCSI: abort rejected (0x%x) for itt %u, rtt %u, task %p, cmnd %p, Cmd 0x%x\n", 
                   ststmrh->response, ntohl(ststmrh->itt), ntohl(ststmrh->rtt), task, task->scsi_cmnd, task->scsi_cmnd->cmnd[0]);
        }
        else {
            printk("iSCSI: abort successful for itt %u, rtt %u, task %p, cmnd %p, Cmd 0x%x\n",
                   ntohl(ststmrh->itt), ntohl(ststmrh->rtt), task, task->scsi_cmnd, task->scsi_cmnd->cmnd[0]);
        }
        ISCSI_TRACE( ISCSI_TRACE_RxAbort, 0, 0, 0, (unsigned int)task, ststmrh->response );
    }        
    else {
        printk("iSCSI: mgmt response 0x%x for unknown itt %u, rtt %u\n", 
               ststmrh->response, ntohl(ststmrh->itt), ntohl(ststmrh->rtt));
    }
    spin_unlock(&session->task_lock);
}


static void iscsi_recv_async_event(iscsi_session_t *session, struct IscsiAsyncEvtHdr *staeh )
{
    updateSN(session, ntohl(staeh->expcmdsn), ntohl(staeh->maxcmdsn));

    ISCSI_TRACE( ISCSI_TRACE_RxAsnEv, 0, staeh->async_event, staeh->async_vcode, ntoh24(staeh->dlength), 0);

    switch (staeh->async_event) {
        case ASYNC_EVENT_SCSI_EVENT:
            DEBUG_ERR0("iSCSI: SCSI async event\n");
            break;
        case ASYNC_EVENT_REQUEST_LOGOUT:
            DEBUG_ERR2("iSCSI: target requests logout within %u seconds for session to %s\n", 
                       ntohs(staeh->param3), session->TargetName);
            /* FIXME: this is really a request to drop a connection, not the whole session,
             * but we currently only have one connection per session, so there's no difference
             * at the moment.
             */
            session->logout_deadline = jiffies + (ntohs(staeh->param3) * HZ);
            set_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits);
            /* FIXME: pass this up to the daemon?  try to unmount
             * filesystems? decide what to do, then do it, and make
             * sure we'll get woken up to send the target a logout.
             * The spec confusingly says initiators MUST send a
             * logout, and then goes on to define what targets should
             * do if the initiator doesn't send a Logout.
             */
            break;
        case ASYNC_EVENT_DROPPING_CONNECTION:
            DEBUG_ERR4("iSCSI: target dropping connection %u for session to %s, reconnect min %u max %u\n", 
                       ntohs(staeh->param1), session->TargetName, ntohs(staeh->param2), ntohs(staeh->param3));
            session->min_reconnect_time = jiffies + (ntohs(staeh->param2) * HZ);
            /* FIXME: obey the min reconnect time */
            break;
        case ASYNC_EVENT_DROPPING_ALL_CONNECTIONS:
            DEBUG_ERR3("iSCSI: target dropping all connections for session to %s, reconnect min %u max %u\n", 
                       session->TargetName, ntohs(staeh->param2), ntohs(staeh->param3));
            session->min_reconnect_time = jiffies + (ntohs(staeh->param2) * HZ);
            /* FIXME: obey the min reconnect time */
            break;
        case ASYNC_EVENT_VENDOR_SPECIFIC:
            DEBUG_ERR2("iSCSI: vendor-specific async event, vcode 0x%x, received on session to %s\n",
                       staeh->async_vcode, session->TargetName);
            break;
        default:
            printk("iSCSI: unknown async event 0x%x received on session to %s\n",
                   staeh->async_event, session->TargetName);
            break;
    }

    return;
}

static void iscsi_rx_thread(iscsi_session_t *session)
{
    int status, rc, xlen, pad;
    struct msghdr msg;
    struct iovec iov;
    struct IscsiHdr sth;
    unsigned char *rxbuf;
    mm_segment_t fs;

    DEBUG_INIT2("iSCSI: rx thread to %s, current %p\n", session->TargetName, current);

    while (!signal_pending(current)) {

        /* check for anything to read on socket */
        iov.iov_base = &sth;
        iov.iov_len = sizeof(sth);
        memset( &msg, 0, sizeof(struct msghdr) );
        msg.msg_iov = &iov;
        msg.msg_iovlen = 1;

        DEBUG_FLOW1("iSCSI: rx thread %s waiting to receive\n", session->TargetName);

        rc = iscsi_recvmsg( session, &msg, sizeof(sth) );
        if (signal_pending(current)) {
            DEBUG_FLOW1("iSCSI: rx thread %s received signal\n", session->TargetName);
            goto ThreadExit;
        }
        if ( rc == sizeof(sth) ) {
            /* FIXME: read any additional header segments as well */

            /* received something */
            xlen = ntoh24( sth.dlength );

            /* If there are padding bytes, read them as well */
            pad = xlen % PAD_WORD_LEN;
            if (pad) {
                pad = PAD_WORD_LEN - pad;
                xlen += pad;
            }

            DEBUG_FLOW3("iSCSI: rx PDU, opcode %x, len %d, pad %d\n", sth.opcode, xlen, pad);
            if ( xlen && (sth.opcode != ISCSI_OP_SCSI_DATA_RSP) ) {
                if ( xlen > ISCSI_RXCTRL_SIZE ) {
                    DEBUG_ERR2("iSCSI: PDU data length too large, opcode %x, dlen %d\n", sth.opcode, xlen);
                    if (session->rx_pid)
                        kill_proc(session->rx_pid, SIGPIPE, 1);
                    break;
                }
                rxbuf = session->RxBuf;
                iov.iov_base = rxbuf;
                iov.iov_len = xlen;
                memset( &msg, 0, sizeof(struct msghdr) );
                msg.msg_iov = &iov;
                msg.msg_iovlen = 1;
                rc = iscsi_recvmsg( session, &msg, xlen );
                if ( rc != xlen ) {
                    DEBUG_ERR3("iSCSI: PDU opcode %x, recvmsg %d failed, %d\n", sth.opcode, xlen, rc);
                    if (session->rx_pid)
                        kill_proc(session->rx_pid, SIGPIPE, 1);
                    break;
                }
            } 
            else {
                rxbuf = NULL;
            }
            
            switch( sth.opcode ) {
                case ISCSI_OP_NOOP_IN:
                    iscsi_recv_nop( session, (struct IscsiNopInHdr *)&sth );
                    break;
                case ISCSI_OP_SCSI_RSP:
                    session->ExpStatSn = ntohl(((struct IscsiScsiRspHdr *)&sth)->statsn)+1;
                    iscsi_recv_cmd(session, (struct IscsiScsiRspHdr *)&sth, rxbuf );
                    break;
                case ISCSI_OP_SCSI_TASK_MGT_RSP:
                    session->ExpStatSn = ntohl(((struct IscsiScsiTaskMgtRspHdr *)&sth)->statsn)+1;
                    iscsi_recv_mgmt_rsp(session, (struct IscsiScsiTaskMgtRspHdr *)&sth );
                    break;
                case ISCSI_OP_RTT_RSP:
                    iscsi_recv_r2t(session, (struct IscsiRttHdr *)&sth );
                    break;
                case ISCSI_OP_SCSI_DATA_RSP:
                    iscsi_recv_data( session, (struct IscsiDataRspHdr *)&sth );
                    break;
                case ISCSI_OP_ASYNC_EVENT:
                    session->ExpStatSn = ntohl(((struct IscsiAsyncEvtHdr *)&sth)->statsn)+1;
                    iscsi_recv_async_event(session, (struct IscsiAsyncEvtHdr *)&sth );
                    break;
                case ISCSI_OP_REJECT_MSG:
                    DEBUG_ERR0("iSCSI: target rejected PDU, dropping connection\n");
                    goto ThreadExit;
                default:
                    DEBUG_ERR1("iSCSI: received unexpected opcode 0x%x, dropping connection\n", sth.opcode);
                    goto ThreadExit;
            }
        } 
        else {
            if ( rc != -EAGAIN ) {
                if (rc == 0) {
                    printk( "iSCSI: Connection to %s closed by target.\n", session->TargetName);
                }
                else if (rc == -ECONNRESET) {
                    printk( "iSCSI: Connection to %s reset.\n", session->TargetName);
                } 
                else if ( rc == -ERESTARTSYS ) {
                    printk( "iSCSI: Connection to %s terminated.\n", session->TargetName);
                } 
                else {
                    printk("iSCSI: Unexpected read status %d on connection to %s\n", rc, session->TargetName);
                }
                goto ThreadExit;
            }
        }
    }
    
 ThreadExit:
    printk("iSCSI: rx thread %s going down\n", session->TargetName);
    /* indicate that we're already going down, so that we don't get killed */
    session->rx_pid = 0;
    mb();

    fs = get_fs();   
    set_fs (get_ds()); 
    while ((rc = waitpid(session->tx_pid, &status, __WCLONE|WNOHANG)) == -ERESTARTSYS) {
        DEBUG_INIT1("iSCSI: restarting waitpid %u\n", session->tx_pid);
    }
    set_fs(fs); 
    
    if (rc == session->tx_pid) {
        session->tx_pid = 0;
    }
    else {
        /* if we can't reap it now, kill it, and try again later */
        DEBUG_INIT2("iSCSI: rx thread %s killing tx_pid %d\n", session->TargetName, session->tx_pid);
        if (session->tx_pid)
            kill_proc(session->tx_pid, SIGKILL, 1);
    }

    /* disconnect a target */
    atomic_dec(&NumTargets);
    
    /* wait for the TX thread to exit. We need a better way of doing
     * this. Perhaps forking in the daemon and using 2 separate ioctls
     * for connection rx and tx, rather than using
     * kernel_thread() and waitpid() from within the kernel module.
     * On the other hand, if we want to get away from ioctl()s that
     * block for long periods of time, we need to use kernel threads.
     */
    if (session->tx_pid) {
        DEBUG_INIT2("iSCSI: rx thread %s waiting for tx_pid %d to exit\n", session->TargetName, session->tx_pid);
        fs = get_fs();     /* save previous value */
        set_fs (get_ds()); /* use kernel limit */
        do {
            rc = waitpid(session->tx_pid, &status, __WCLONE);
        } while (rc == -ERESTARTSYS);
        set_fs(fs); /* restore before returning to user space */
        if (rc > 0) {
            DEBUG_INIT2("iSCSI: reaped tx_pid %d, status 0x%x\n", session->tx_pid, status);
            session->tx_pid = 0;
        }
        else {
            printk("iSCSI: failed to reap tx_pid %d, rc %d, errno %d\n", session->tx_pid, rc, errno);
        }
    }

    set_current_state(TASK_RUNNING);
    DEBUG_INIT1("iSCSI: rx thread for %s done\n", session->TargetName);
}


static int
iscsi_session( iscsi_session_ioctl_t *ioctld )
{
    struct file *ctrlfile;
    struct inode *ctrlinode;
    iscsi_session_t *session;
    iscsi_hba_t *hba;
    struct socket *sock;
    int hba_number;
    int channel_number;
    iscsi_task_t *task;
    Scsi_Cmnd *sc = NULL;
    Scsi_Cmnd *unsent_cmnds = NULL;
    int result = HOST_BYTE(DID_SOFT_ERROR);
    DECLARE_CPU_FLAGS;
    
    DEBUG_INIT1("iSCSI: ioctl for session to %s\n", ioctld->TargetName);

    ctrlfile = fget( ioctld->socket_fd );
    if ( ! ctrlfile ) {
        return -EINVAL;
    }
    
    ctrlinode = ctrlfile->f_dentry->d_inode;
    if ( !ctrlinode || !ctrlinode->i_sock ) {
        fput( ctrlfile );
        return -EINVAL;
    }
    
    sock = &ctrlinode->u.socket_i;
    if ( sock->file != ctrlfile ) {
        fput( ctrlfile );
        return -EINVAL;
    }

    sock->sk->allocation = GFP_ATOMIC;
    
    /* check to see if this session should be accepted */

    /* find the HBA that has the requested iSCSI bus */
    hba_number = ioctld->iscsi_bus / ISCSI_MAX_CHANNELS_PER_HBA;
    channel_number = ioctld->iscsi_bus % ISCSI_MAX_CHANNELS_PER_HBA;

    spin_lock(&hba_list_lock);
    hba = hba_list;
    while (hba && (hba_number-- > 0)) {
        hba = hba->next;
    }
    if (!hba) {
        spin_unlock(&hba_list_lock);
        printk("iSCSI: couldn't find HBA with iSCSI bus %d\n", ioctld->iscsi_bus);
        return -EINVAL;
    }
    spin_unlock(&hba_list_lock);

    /* check sessions on the HBA */
    spin_lock_irqsave(&hba->session_lock, cpu_flags);
    session = hba->session_list_head;
    while (session) {
        if ( session->address_length == ioctld->AddressLength &&
             memcmp(session->ip_address, ioctld->RemoteIpAddr, session->address_length) == 0 &&
             strcmp(session->TargetName, ioctld->TargetName) == 0) {
            if (session->channel == channel_number &&
                session->target_id == ioctld->TargetId) {
                if (session->rx_pid) {
                    /* already have a session running to this target */
                    printk("iSCSI: session to %s already exists\n", ioctld->TargetName);
                    spin_unlock_irqrestore(&hba->session_lock, cpu_flags);
                    return -EINVAL;
                }
                else {
                    /* same iSCSI bus and target id, no current session, it's a valid session restart */
                    break;
                }
            }
            else {
                printk("iSCSI: error - TargetName %s cannot move from bus %d id %d to bus %d id %d\n", ioctld->TargetName, 
                       session->iscsi_bus, session->target_id, ioctld->iscsi_bus, ioctld->TargetId);
                fput(ctrlfile);
                spin_unlock_irqrestore(&hba->session_lock, cpu_flags);
                return -EINVAL;
            }
        }
        else if (session->channel == channel_number &&
                 session->target_id == ioctld->TargetId) {
            printk("iSCSI: error - TargetName %s cannot claim bus %d id %d, already in use by %s\n", 
                   ioctld->TargetName, ioctld->iscsi_bus, session->target_id, session->TargetName);
            fput(ctrlfile);
            spin_unlock_irqrestore(&hba->session_lock, cpu_flags);
            return -EPERM;
        }
        session = session->next;
    }
    
    if ( session ) {
        session->generation++;
        printk("iSCSI: session #%lu to %s entering full-feature phase\n", 
               session->generation, ioctld->TargetName);

        clear_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits);
        clear_bit(SESSION_WINDOW_CLOSED, &session->control_bits);

        /* make sure we start sending commands again */
        set_bit(TX_SCSI_COMMAND, &session->control_bits);
    } 
    else {
        /*
         * must be a new session
         */
        printk("iSCSI: session #1 to %s entering full-feature phase\n", ioctld->TargetName);

        session = (iscsi_session_t *)kmalloc( sizeof(iscsi_session_t), GFP_KERNEL );
        if ( ! session ) {
            DEBUG_ERR0("iSCSI: Cannot allocate new session entry\n");
            fput( ctrlfile );
            return -ENOMEM;
        }
        memset( session, 0, sizeof(iscsi_session_t) );

        session->hba = hba;
        session->generation = 1;

        spin_lock_init( &session->scsi_cmnd_lock);
        session->scsi_cmnd_head = session->scsi_cmnd_tail = NULL;

        spin_lock_init( &session->task_lock);
        session->arrival_order.head = session->arrival_order.tail = NULL;
        session->rx_tasks.head = session->rx_tasks.tail = NULL;
        session->tx_tasks.head = session->tx_tasks.tail = NULL;
        session->rx_abort_tasks.head = session->rx_abort_tasks.tail = NULL;
        session->tx_abort_tasks.head = session->tx_abort_tasks.tail = NULL;
        session->aborted_tasks.head = session->aborted_tasks.tail = NULL;
        session->completing_tasks.head = session->completing_tasks.tail = NULL;
        session->tx_lun_reset_tasks.head = session->tx_lun_reset_tasks.tail = NULL;
        atomic_set(&session->num_tasks, 0);

#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
        init_waitqueue_head(&session->tx_wait_q);
#else
        session->tx_wait_q = NULL;
#endif
        /* copy the IP address and port for the /proc entries */
        session->address_length = ioctld->AddressLength;
        memcpy(session->ip_address, ioctld->RemoteIpAddr, ioctld->AddressLength);
        session->Port = ioctld->Port;

        /* record target name, id */
        memset(session->TargetName, 0, sizeof(session->TargetName));
        strncpy(session->TargetName, ioctld->TargetName, sizeof(session->TargetName));
        session->TargetName[TARGET_NAME_MAXLEN] = '\0';
        session->target_id = ioctld->TargetId;
        session->iscsi_bus = ioctld->iscsi_bus;
        session->channel = ioctld->iscsi_bus % ISCSI_MAX_CHANNELS_PER_HBA;

        /* init ITT */
        session->itt = 1;

        /* attach the session to the HBA */
        if (hba->session_list_head) {
            session->next = NULL;
            session->prev = hba->session_list_tail;
            hba->session_list_tail->next = session;
            hba->session_list_tail = session;
        }
        else {
            hba->session_list_head = hba->session_list_tail = session;
            session->next = session->prev = NULL;
        }        
    }
    spin_unlock_irqrestore(&hba->session_lock, cpu_flags);

    atomic_inc(&NumTargets);
    DEBUG_INIT1("iSCSI: Target count: %d\n", atomic_read(&NumTargets));

    session->socket = sock;
    session->CmdSn = ioctld->CmdSn;
    session->ExpCmdSn = ioctld->ExpCmdSn;
    session->MaxCmdSn = ioctld->MaxCmdSn;
    session->ExpStatSn = ioctld->InitStatSn;
    session->initialR2T = ioctld->initialR2T;
    session->dataPDULength = ioctld->dataPDULength << 9; /* convert from 512 byte units */
    session->firstBurstSize= ioctld->firstBurstSize << 9; /* convert from 512 byte units */
    session->maxBurstSize= ioctld->maxBurstSize << 9; /* convert from 512 byte units */
    session->immediateData = ioctld->immediateData;
    session->ActiveTimeout = ioctld->ActiveTimeout;
    session->IdleTimeout = ioctld->IdleTimeout;
    session->PingTimeout = ioctld->PingTimeout; 
    session->AbortTimeout = ioctld->AbortTimeout;
    session->ResetTimeout = ioctld->ResetTimeout;
    session->last_rx = jiffies;
    session->last_ping = jiffies - 1;
    session->last_window_check = jiffies;
    session->warm_reset_itt = RSVD_TASK_TAG;
    session->cold_reset_itt = RSVD_TASK_TAG;

    /* queue up a ping */
    set_bit(TX_PING, &session->control_bits);

    while (! TimerActive ) {
        DEBUG_INIT1("iSCSI: iscsi_session %s waiting for timer thread\n", session->TargetName);
        if ( signal_pending(current) ) {
            break;
        }
        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout( 10 );
    }

    DEBUG_INIT1("iSCSI: iscsi_session %s about to start tx and rx threads\n", 
                session->TargetName);
    
    session->tx_pid = kernel_thread( iscsi_tx_thread, (void *)session,
                                     (CLONE_VM|CLONE_FS|CLONE_FILES|CLONE_SIGHAND) );

    /* become the rx thread */
    session->rx_pid = current->pid;
    iscsi_rx_thread(session);

    /* session is now over */

    /* get the io request lock, so we can call the Scsi_Cmnds' done function */
    LOCK_IO_REQUEST_LOCK;

    /* grab the unsent commands already queued, to ensure
     * they get aborted later.  This ensures dependent SCSI commands
     * won't get re-ordered, in cases where the upcoming completion
     * re-queues a command behind these unsent commands.  This may
     * be unneccesary, if the mid-layer or high-level driver 
     * is already handling dependent commands.
     */
    spin_lock(&session->scsi_cmnd_lock);
    unsent_cmnds = session->scsi_cmnd_head;
    session->scsi_cmnd_head = session->scsi_cmnd_tail = NULL;
    spin_unlock(&session->scsi_cmnd_lock);

    /* abort all sent commands (tasks) for this connection */
    spin_lock(&session->task_lock);

    if (session->cold_reset_itt != RSVD_TASK_TAG) {
        result = HOST_BYTE(DID_RESET);
        printk("iSCSI: target reset, aborting SCSI commands for session %p to %s with DID_RESET\n", 
               session, session->TargetName);
    }
    else {
        result = HOST_BYTE(DID_SOFT_ERROR);
        printk("iSCSI: session dropped, aborting SCSI commands for session %p to %s with DID_SOFT_ERROR\n", 
               session, session->TargetName);
    }

    /* clear out the task collections */
    session->rx_tasks.head = session->rx_tasks.tail = NULL;
    session->tx_tasks.head = session->tx_tasks.tail = NULL;
    session->rx_abort_tasks.head = session->rx_abort_tasks.tail = NULL;
    session->tx_abort_tasks.head = session->tx_abort_tasks.tail = NULL;
    session->aborted_tasks.head = session->aborted_tasks.tail = NULL;
    session->completing_tasks.head = session->completing_tasks.tail = NULL;
    session->tx_lun_reset_tasks.head = session->tx_lun_reset_tasks.tail = NULL;
    session->warm_reset_itt = RSVD_TASK_TAG;
    session->cold_reset_itt = RSVD_TASK_TAG;
    /* tasks that are aborting have already been removed from the arrival_order collection,
     * so we can just complete everything in it.
     */
    while ((task = session->arrival_order.head)) {
        DEBUG_QUEUE2("iSCSI: aborting sent task %p, command %p\n", task, task->scsi_cmnd);
        session->arrival_order.head = task->order_next;
        sc = task->scsi_cmnd;
        task->scsi_cmnd = NULL;
        free_task(task);
        if (sc && sc->scsi_done) {
            sc->host_scribble = NULL;
            sc->result = result;
            sc->scsi_done(sc);
            printk("iSCSI: aborted sent task %p, command %p\n", task, sc);
        }
    }
    session->arrival_order.head = session->arrival_order.tail = NULL;
    atomic_set(&session->num_tasks, 0);
    spin_unlock(&session->task_lock);
    
    while ((sc = unsent_cmnds)) {
        unsent_cmnds = (Scsi_Cmnd *)sc->host_scribble;
        DEBUG_QUEUE1("iSCSI: aborting unsent SCSI command %p\n", sc);
        if ( sc->scsi_done) {
            sc->host_scribble = NULL;
            sc->result = result;
            sc->scsi_done(sc);
            printk("iSCSI: aborted unsent SCSI command %p\n", sc);
        }
        else {
            printk("iSCSI: unsent command %p already aborting\n", sc);
        }
    }

    printk("iSCSI: done aborting SCSI commands for session %p to %s\n", session, session->TargetName);

    UNLOCK_IO_REQUEST_LOCK;

    if (result == HOST_BYTE(DID_RESET)) {
        /* inform the eh_*_reset functions that the reset is completed */
        if (test_bit(SESSION_RESETTING, &session->control_bits)) {
            set_bit(SESSION_RESET, &session->control_bits);
            clear_bit(SESSION_RESETTING, &session->control_bits);
        }
    }

    /* ioctl finally returning out of the kernel */
    printk("iSCSI: session to %s leaving kernel\n", ioctld->TargetName);

    /* FIXME: do we really want to do this?  If we can keep the socket open
     * after the daemon dies, perhaps we could leave iSCSI disks mounted
     * until the system would normally unmount and sync devices prior to
     * shutdown/reboot, rather than have to do it ourselves in the scripts.
     */
    fput( ctrlfile );

    return 0;
}


/*
 *  FIXME: it'd probably be cleaner to move the timeout logic to the rx thread.
 *         The only danger is if the rx thread somehow blocks indefinately.
 *         Doing timeouts here makes sure the timeouts get checked, at the
 *         cost of having this code constantly loop.
 */
static int
iscsi_timer_thread( void ) { 
    int rc; 
    iscsi_session_t *session;
    iscsi_hba_t *hba;

    DEBUG_INIT0("iSCSI: Timer\n");

    TimerActive = 1;

    /* wait for an HBA to exist */
    spin_lock(&hba_list_lock);
    while (hba_list == NULL) {
        spin_unlock(&hba_list_lock);
        if ( signal_pending(current) ) {
            return 0;
        }
        set_current_state(TASK_INTERRUPTIBLE);
        rc = schedule_timeout( 10 );
        DEBUG_FLOW0("iSCSI: timer running\n");
        spin_lock(&hba_list_lock);
    }

    /* FIXME: handle multiple HBAs */
    while ( hba_list ) {
        /* FIXME: consider moving timeout processing to the rx and tx threads, since any timeout
         *        must either wake the tx thread to send a Nop/poll, or kill the session
         *        and all it's threads.  Either way, we need at least one of the threads awake, and so
         *        handling timeouts there lets us get rid of the Timer thread.
         */
        spin_unlock(&hba_list_lock);

        set_current_state(TASK_INTERRUPTIBLE);
        /* FIXME: schedule a timeout for the earliest possible session timeout, so we that we sleep longer */
        schedule_timeout(10);
        if ( signal_pending(current) ) {
            DEBUG_INIT0("iSCSI: Timer leaving kernel\n");
            TimerActive = 0;
            set_current_state(TASK_RUNNING);
            return 0;
        }

        spin_lock(&hba_list_lock);
        hba = hba_list;
        while (hba) {
            cpu_flags_t cpu_flags;

            /* process timeouts for all sessions */
            spin_lock_irqsave(&hba->session_lock, cpu_flags);
            session = hba->session_list_head;
            while ( session ) {
                unsigned long timeout;

                if (session->rx_pid) {
                    if (test_bit(SESSION_WINDOW_CLOSED, &session->control_bits) &&
                        jiffy_lt(session->last_window_check + HZ, jiffies)) {
                        /* command window closed, ping once a second to ensure we find out
                         * when it re-opens.  Target should send us an update when it does,
                         * but we're not very trusting of target correctness.
                         */
                        session->last_window_check = jiffies;
                        wake_tx_thread(TX_PING, session);
                        printk("iSCSI: session %s command window closed, ExpCmdSN %u, MaxCmdSN %u\n", 
                               session->TargetName, session->ExpCmdSn, session->MaxCmdSn);
                    }
                    
                    if (atomic_read(&session->num_tasks))
                        timeout = session->ActiveTimeout;
                    else
                        timeout = session->IdleTimeout;
                    
                    if (timeout) {
                        if (session->PingTimeout &&
                            jiffy_lt(session->last_rx + (timeout * HZ) + (session->PingTimeout * HZ), jiffies)) {
                            /* should have received something by now, kill the connection */
                            printk("iSCSI: dropping session %s, %lu second timeout expired, rx %lu, ping %lu\n", 
                                   session->TargetName, timeout + session->PingTimeout, session->last_rx, session->last_ping);
                            if ( session->tx_pid ) {
                                printk("iSCSI: killing tx thread %d, session %p to %s\n", session->tx_pid, session, session->TargetName);
                                kill_proc(session->tx_pid, SIGKILL, 1);
                            }
                            if ( session->rx_pid ) {
                                printk("iSCSI: killing rx thread %d, session %p to %s\n", session->rx_pid, session, session->TargetName);
                                kill_proc(session->rx_pid, SIGPIPE, 1);
                            }
                        }
                        else if (jiffy_lt(session->last_rx + (timeout * HZ), jiffies) &&
                                 jiffy_lt(session->last_ping, session->last_rx)) {
                            /* send a ping to try to provoke some traffic */
                            DEBUG_FLOW3("iSCSI: sending ping to %s, rx %lu, ping %lu\n", 
                                    session->TargetName, session->last_rx, session->last_ping);
                            session->last_ping = jiffies - 1;
                            wake_tx_thread(TX_PING, session);
                        }
                    }
                }
                session = session->next;
            }
            spin_unlock_irqrestore(&hba->session_lock, cpu_flags);
            hba = hba->next;
        }
    }

    /* timer finished */
    DEBUG_INIT0("iSCSI: Timer leaving kernel\n");

    spin_unlock(&hba_list_lock);
    
    TimerActive = 0;
    
    set_current_state(TASK_RUNNING);
    
    return 0;
}


int
iscsi_detect( Scsi_Host_Template *sht )
{
    struct Scsi_Host *sh;
    iscsi_hba_t *hba;
    iscsi_task_t *task;
    int i; 

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
    sht->proc_dir = &proc_dir_iscsi;
#else
    sht->proc_name = "iscsi";
#endif

    sh = scsi_register( sht, sizeof(iscsi_hba_t) );
    if (!sh ) {
        printk("iSCSI: Unable to register controller\n");
        return 0;
    }

    sh->max_id = ISCSI_MAX_TARGETS;
    sh->max_lun = ISCSI_MAX_LUN;
    sh->max_channel = ISCSI_MAX_CHANNELS_PER_HBA - 1; /* convert from count to index */

    hba = (iscsi_hba_t *)sh->hostdata;
    memset( hba, 0, sizeof(iscsi_hba_t) );

    hba->next = NULL;

    /* record our Linux host (HBA) number for later use */
    hba->hostno = sh->host_no;

    /* list of sessions on this HBA */
    spin_lock_init(&hba->session_lock);
    hba->session_list_head = NULL;
    hba->session_list_tail = NULL;

    /* pre-allocate command entries for this HBA 
     * FIXME: this doesn't scale.  We ought to be allocating these
     * per session, or better yet, per LUN detected.
     */
    spin_lock_init( &hba->task_lock);
    i = (ISCSI_TASKS_PER_HBA * sizeof(iscsi_task_t));
    task = kmalloc( i, GFP_KERNEL);
    if ( ! task ) {
        printk("iSCSI: Detect could not allocate memory\n");
        return 0;
    }
    memset(task, 0, i);
    task->mgmt_itt = RSVD_TASK_TAG;
    task->itt = RSVD_TASK_TAG;
    hba->free_task_head = hba->task_mem = task;
    hba->free_task_tail = hba->task_max = task + ISCSI_TASKS_PER_HBA - 1;
    task->prev = NULL;
    while (task < hba->task_max) {
        task->mgmt_itt = RSVD_TASK_TAG;
        task->itt = RSVD_TASK_TAG;
        task->next = (task + 1);
        task->next->prev = task;
        task++;
    }
    task->next = NULL;

    hba->active = 1;

    /* for now, there's just one iSCSI HBA */
    hba_list = hba;
    printk("iSCSI: detected HBA %p, host #%d\n", hba, hba->hostno);
    return 1;
}

int
iscsi_release( struct Scsi_Host *sh )
{
    iscsi_hba_t *hba;
    iscsi_session_t *session;

    hba = (iscsi_hba_t *)sh->hostdata;
    if ( ! hba ) {
        return FALSE;
    }

    printk("iSCSI: release HBA %p, host #%d\n", hba, hba->hostno);
    
    /* remove all sessions on this HBA */
    spin_lock(&hba->session_lock);
    session = hba->session_list_head;
    while (session) {
        /* FIXME: wait for all session tasks to complete, or abort them? */
        /* FIXME: abort all queued SCSI commands for this session (or just let them timeout in the mid-layer)?  */
        /* probably unneccsary, since the daemon will keep the module use count non-zero, so the module
         * can't possibly unload until after the daemon dies, and all the sessions are gone by then.
         */
        hba->session_list_head = session->next;
        kfree( session );
        session = hba->session_list_head;
    }
    spin_unlock(&hba->session_lock);

    /* free all the task memory */
    kfree( hba->task_mem );
    
    /* remove from the iSCSI HBA list */
    spin_lock(&hba_list_lock);
    if (hba == hba_list) {
        hba_list = hba_list->next;
    }
    else {
        iscsi_hba_t *prior = hba_list;

        while (prior && prior->next != hba)
            prior = prior->next;
        if (prior && prior->next == hba)
            prior->next = hba->next;
    }
    spin_unlock(&hba_list_lock);

    memset( hba, 0, sizeof(iscsi_hba_t) );
    scsi_unregister( sh );

    return FAILED;
}

#ifdef ISCSI_USE_STRATEGY
int
iscsi_strategy( struct Scsi_Host *host )
{
    Scsi_Device *sd;
    Scsi_Cmnd *sc;
    iscsi_hba_t *hba = (iscsi_hba_t *)host->hostdata;
    iscsi_session_t *session;
    iscsi_task_t *task;

    for( sd=host->host_queue; sd; sd=sd->next ) {
        printk("iSCSI: Strategy Device 0x%x, Id %d, Lun %d, Chan %d\n",(int)sd,sd->id,sd->lun,sd->channel);
        for( sc=sd->device_queue; sc; sc=sc->next ) {
            printk("iSCSI: Strategy Cmd 0x%x, state 0x%x, tgt 0x%x\n",(int)sc,sc->state,sc->target);
        }
    }

    spin_lock(&hba->session_lock);
    session = hba->session_list_head;
    while( session ) {
        sc = session->scsi_cmnd_head;
        while( sc ) {
            printk("iSCSI: scsi_cmnd_head 0x%x\n",(int)sc);
            sc = (Scsi_Cmnd *)sc->host_scribble;
        }
        task = session->task_head;
        while( task ) {
            printk("iSCSI: task_head 0x%x\n",(int)task->sc);
            task = task->next;
        }
        session = session->next;
    }
    spin_unlock(&hba->session_lock);

    return 0;
}
#endif


int
iscsi_eh_abort( Scsi_Cmnd *sc )
{
    struct Scsi_Host *host = NULL;
    iscsi_hba_t *hba = NULL;
    iscsi_task_t *task = NULL;
    iscsi_session_t *session = NULL;
    int completed = 0;
    int refcount, ret = FAILED;
    unsigned long timeout;
    unsigned long interval = (HZ / 10) ? (HZ / 10) : 10;
    unsigned long last_log = 0;
    unsigned long current_gen;

    if ( ! sc ) {
        return FAILED;
    }
    host = sc->host;
    if (! host) {
        printk("iSCSI: no host for SCSI command %p\n", sc);
        return FAILED;
    }
    hba = (iscsi_hba_t *)host->hostdata;
    if (!hba) {
        printk("iSCSI: no iSCSI HBA associated with SCSI command %p\n", sc);
        return FAILED;
    }

    /* find the appropriate session for the command */
    session = find_session_for_cmnd(sc);
    if (!session) {
        printk("iSCSI: can't abort cmnd %p, no session for command\n", sc);
        return FAILED;
    }

    sc->result = HOST_BYTE(DID_ABORT);

    RELEASE_IO_REQUEST_LOCK;

    printk("iSCSI: eh_abort for command %p to (%u %u %u %u), Cmd 0x%x\n",
           sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);

    /* check if it's a command we haven't sent yet (still in the scsi_cmnd queue).
     * if we find the command, remove it from the queue.
     */
    spin_lock(&session->scsi_cmnd_lock);
    if (sc == session->scsi_cmnd_head) {
        /* it's the head, remove it */
        session->scsi_cmnd_head = (Scsi_Cmnd *)sc->host_scribble; /* next */
        if (sc == session->scsi_cmnd_tail)
            session->scsi_cmnd_tail = NULL;
        printk("iSCSI: aborted unsent command %p to (%u %u %u %u), Cmd 0x%x\n",
               sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        spin_unlock(&session->scsi_cmnd_lock);
        ret = SUCCESS;
        goto done;
    }
    else if (session->scsi_cmnd_head) {
        Scsi_Cmnd *prior, *next;
        
        /* find the command prior to sc */
        prior = session->scsi_cmnd_head;
        next = (Scsi_Cmnd *)session->scsi_cmnd_head->host_scribble;
        while (next && next != sc) {
            prior = next;
            next = (Scsi_Cmnd *)prior->host_scribble; /* next command */
        }
        if (prior && next == sc) {
            /* remove the command */
            prior->host_scribble = next->host_scribble;
            if (session->scsi_cmnd_tail == sc)
                session->scsi_cmnd_tail = prior;
            sc->host_scribble = NULL;
            printk("iSCSI: aborted unsent command %p to (%u %u %u %u), Cmd 0x%x\n",
                   sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
            spin_unlock(&session->scsi_cmnd_lock);
            ret = SUCCESS;
            goto done;
        }
    }
    spin_unlock(&session->scsi_cmnd_lock);

    /* 
     * check tasks (commands already sent to the target) 
     */
    spin_lock(&session->task_lock);
    current_gen = session->generation;
    if ((task = remove_task_for_cmnd(&session->rx_tasks, sc))) {
        /* it's waiting for data, an R2T, or a command response */
        task->mgmt_itt = 0;
        atomic_inc(&task->refcount);
        add_task(&session->tx_abort_tasks, task);
        /* need to half-remove it from the session to make sure it won't be completed if the session drops */
        remove_session_task(session, task);
    }
    else if ((task = remove_task_for_cmnd(&session->tx_tasks, sc))) {
        /* it's received an R2T, and is queued to have data sent */
        task->mgmt_itt = 0;
        atomic_inc(&task->refcount);
        add_task(&session->tx_abort_tasks, task);
        /* need to half-remove it from the session to make sure it won't be completed if the session drops */
        remove_session_task(session, task);
    }
    else if ((task = remove_task_for_cmnd(&session->completing_tasks, sc))) {
        /* already received command completion, no point in sending an abort */
        completed = 1;
        task->mgmt_itt = 0;
        atomic_inc(&task->refcount);
        DEBUG_INIT3("iSCSI: eh_abort - no need to send abort for itt %u, task %p, cmnd %p completed\n",
                    task->itt, task, sc);
        /* need to half-remove it from the session to make sure it won't be completed if the session drops */
        remove_session_task(session, task);
    }
    spin_unlock(&session->task_lock);
    
    if (task) {

        if (!completed) {
            iscsi_task_t *aborted_task = NULL;
            
            printk("iSCSI: aborting itt %u, task %p, command %p to (%u %u %u %u), Cmd 0x%x\n",
                   task->itt, task, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
            wake_tx_thread(TX_ABORT, session);
            
            /* wait for an abort response for the abort we just queued up
             * to be sent.  it's possible the target will never respond,
             * so after a configurable timeout, just wait for the task's
             * refcount to hit zero, and then free the task.  It's
             * possible the refcount won't hit zero, if the tx thread or
             * rx thread is blocked trying to read from or write to the
             * command's buffer, but in that case the timer thread will
             * eventually kill the session, and we can return as soon as
             * the session starts going down.  
             */
            timeout = jiffies + (session->AbortTimeout * HZ);
            /* until the session drops or the timeout expires */
            while (session->rx_pid && (session->generation == current_gen) && jiffy_lt(jiffies, timeout)) {

                spin_lock(&session->task_lock);
                aborted_task = remove_task_for_cmnd(&session->aborted_tasks, sc);
                spin_unlock(&session->task_lock);

                if (aborted_task)
                    break;
                
                /* wait a bit and check again */
                set_current_state(TASK_INTERRUPTIBLE);
                schedule_timeout(interval);
            }

            /* if we still have a session up, and didn't receive an abort by now, quit waiting for it */
            if (session->rx_pid && (session->generation == current_gen) && !aborted_task) {
                spin_lock(&session->task_lock);
                if ((aborted_task = remove_task_for_cmnd(&session->rx_abort_tasks, sc))) {
                    printk("iSCSI: failed to recv abort response for itt %u, task %p, cmnd %p\n",
                           task->itt, task, sc);
                }
                else if ((aborted_task = remove_task_for_cmnd(&session->tx_abort_tasks, sc))) {
                    printk("iSCSI: failed to send abort for itt %u, task %p, cmnd %p\n",
                           task->itt, task, sc);
                }
                spin_unlock(&session->task_lock);
            }
        }
        
        /* wait for the task refcount to be 1 (just us), and then free
         * it.  this should happen eventually, since if nothing else
         * the timer thread will kill the tx and rx threads, and
         * they'll quit using the task then.
         */
        while (task) {
            if ((refcount = atomic_read(&task->refcount)) <= 1) {
#if TEST_RESETS
                /* ignore every 4th successful abort, to test the other error handlers */
                if (session->rx_pid && sc &&
                    (sc->cmnd[0] != TEST_UNIT_READY) &&
                    (task->cmdsn % (RESET_FREQUENCY * ABORT_FREQUENCY)) == 0) {
                    printk("iSCSI: ignoring successful abort of command %p, task %p, (%u %u %u %u), Cmd 0x%x\n",
                           sc, task, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);

                    atomic_dec(&task->refcount);
                    ret = FAILED;
                    break;
                }
#endif
                printk("iSCSI: aborted command %p, task %p, (%u %u %u %u), Cmd 0x%x\n",
                       sc, task, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
                atomic_dec(&task->refcount);
                /* we've already removed the task from the session's ordering collection,
                 * decrement the session's num_tasks to finish doing the moral
                 * equivalent of remove_task_from_session(), then free it.
                 */
                atomic_dec(&session->num_tasks);
                free_task(task);
                task = NULL;
                ret = SUCCESS;
                break;
            }
            
            /* log once per second indicating that eh_abort is waiting */
            if ((last_log == 0) || jiffy_lt(last_log + HZ, jiffies)) {
                last_log = jiffies;
                DEBUG_ERR3("iSCSI: eh_abort waiting for task %p, command %p, refcount %u before returning\n",
                           task, sc, refcount);
            }
        }
    }
    else {
        ret = FAILED;
    }

 done:
    REACQUIRE_IO_REQUEST_LOCK;
    if (ret != SUCCESS) {
        if (task)
            printk("iSCSI: abort failed for command %p, task %p, (%u %u %u %u), Cmd 0x%x\n",
                   sc, task, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        else
            printk("iSCSI: abort failed for command %p, task not found, (%u %u %u %u), Cmd 0x%x\n",
                   sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
    }
    return ret;
}

/* trigger a warm reset */
int warm_reset_target(iscsi_session_t *session)
{
    unsigned long timeout = jiffies + (session->ResetTimeout * HZ);
    unsigned long interval = (HZ / 10) ? (HZ / 10) : 10;
    unsigned long last_log = 0;
    DECLARE_CPU_FLAGS;

    
    /* make sure there is only one oustanding reset, 
     * and prevent queued commands from being sent.
     */
    if (test_and_set_bit(SESSION_RESETTING, &session->control_bits)) {
        /* already resetting */
        printk("iSCSI: session %p to %s already resetting, reset_target failed\n",
               session, session->TargetName);
        return 0;
    }

    /* queue up a reset for the target */
    wake_tx_thread(TX_WARM_TARGET_RESET, session);

    /* wait for it to get sent */
    while (jiffy_lt(jiffies, timeout) && test_bit(TX_WARM_TARGET_RESET, &session->control_bits)) {
        /* log once per second indicating that we're waiting */
        if ((last_log == 0) || jiffy_lt(last_log + HZ, jiffies)) {
            last_log = jiffies;
            printk("iSCSI: waiting for target reset to be sent for session %p to %s\n", 
                   session, session->TargetName);
        }
        /* wait for a bit */
        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(interval);
    }
    
    /* now wait for the reset to actually occur. */
    while (jiffy_lt(jiffies, timeout) && 
           test_bit(SESSION_RESETTING, &session->control_bits) &&
           !test_bit(SESSION_RESET, &session->control_bits)) {
        
        /* log once per second indicating that we're waiting */
        if ((last_log == 0) || jiffy_lt(last_log + HZ, jiffies)) {
            last_log = jiffies;
            printk("iSCSI: waiting for target reset to occur for session %p to %s\n", session, session->TargetName);
        }
        /* wait for a while and check again */
        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(interval);
    }
    
    if (test_and_clear_bit(SESSION_RESET, &session->control_bits)) {
        iscsi_task_t *task;
        Scsi_Cmnd *sc;
        
        printk("iSCSI: aborting commands after target reset for session %p to %s\n", session, session->TargetName);
        
        LOCK_IO_REQUEST_LOCK;

        spin_lock(&session->task_lock);
        /* tasks that are aborting have already been removed from the arrival_order collection,
         * so we can just complete everything in it.
         */
        while ((task = session->arrival_order.head)) {
            printk("iSCSI: aborting sent task %p, command %p\n", task, task->scsi_cmnd);
            session->arrival_order.head = task->order_next;
            sc = task->scsi_cmnd;
            task->scsi_cmnd = NULL;
            free_task(task);
            if (sc && sc->scsi_done) {
                sc->host_scribble = NULL;
                sc->result = HOST_BYTE(DID_RESET);
                printk("iSCSI: aborted sent task %p, command %p\n", task, sc);
                sc->scsi_done(sc);
            }
        }
        session->arrival_order.head = session->arrival_order.tail = NULL;
        atomic_set(&session->num_tasks, 0);
        spin_unlock(&session->task_lock);

        UNLOCK_IO_REQUEST_LOCK;

        printk("iSCSI: target reset succeeded for session %p to %s\n", session, session->TargetName);
        return 1;
    }

    if (jiffy_lt(jiffies, timeout))
        printk("iSCSI: target reset failed for session %p to %s\n", session, session->TargetName);
    else
        printk("iSCSI: target reset timed out for session %p to %s\n", session, session->TargetName);

    /* give up */
    clear_bit(TX_WARM_TARGET_RESET, &session->control_bits);
    if (test_and_clear_bit(SESSION_RESETTING, &session->control_bits)) {
        /* since sending tasks is disabled while resetting, make sure we
         * start sending them again when we stop trying to reset.
         */
        wake_tx_thread(TX_SCSI_COMMAND, session);
    }
    clear_bit(SESSION_RESET, &session->control_bits);
    return 0;
}

/* trigger a cold reset */
int cold_reset_target(iscsi_session_t *session)
{
    unsigned long timeout = jiffies + (session->ResetTimeout * HZ);
    unsigned long interval = (HZ / 10) ? (HZ / 10) : 10;
    unsigned long last_log = 0;

    /* make sure there is only one oustanding reset, 
     * and prevent queued commands from being sent.
     */
    if (test_and_set_bit(SESSION_RESETTING, &session->control_bits)) {
        /* already resetting */
        printk("iSCSI: session %p to %s already resetting, reset_target failed\n",
               session, session->TargetName);
        return 0;
    }

    /* queue up a reset for the target */
    wake_tx_thread(TX_COLD_TARGET_RESET, session);

    /* wait for it to get sent */
    while (jiffy_lt(jiffies, timeout) && test_bit(TX_COLD_TARGET_RESET, &session->control_bits)) {
        /* log once per second indicating that we're waiting */
        if ((last_log == 0) || jiffy_lt(last_log + HZ, jiffies)) {
            last_log = jiffies;
            printk("iSCSI: waiting for cold target reset to be sent for session %p to %s\n", 
                   session, session->TargetName);
        }
        /* wait for a bit */
        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(interval);
    }
    
    /* now wait for the reset to actually occur. */
    while (jiffy_lt(jiffies, timeout) && 
           test_bit(SESSION_RESETTING, &session->control_bits) &&
           !test_bit(SESSION_RESET, &session->control_bits)) {
        
        /* log once per second indicating that we're waiting */
        if ((last_log == 0) || jiffy_lt(last_log + HZ, jiffies)) {
            last_log = jiffies;
            printk("iSCSI: waiting for cold target reset to occur for session %p to %s\n", 
                   session, session->TargetName);
        }
        /* wait for a while and check again */
        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(interval);
    }
    
    if (test_and_clear_bit(SESSION_RESET, &session->control_bits)) {
        /* the commands will have been completed with DID_RESET when the session dropped */
        printk("iSCSI: target reset succeeded for session %p to %s\n", session, session->TargetName);
        return 1;
    }

    if (jiffy_lt(jiffies, timeout))
        printk("iSCSI: cold target reset failed for session %p to %s\n", session, session->TargetName);
    else
        printk("iSCSI: cold target reset timed out for session %p to %s\n", session, session->TargetName);

    /* give up */
    clear_bit(TX_COLD_TARGET_RESET, &session->control_bits);
    if (test_and_clear_bit(SESSION_RESETTING, &session->control_bits)) {
        /* since sending tasks is disabled while resetting, make sure we
         * start sending them again when we stop trying to reset.
         */
        wake_tx_thread(TX_SCSI_COMMAND, session);
    }
    clear_bit(SESSION_RESET, &session->control_bits);
    return 0;
}


int
iscsi_eh_device_reset( Scsi_Cmnd *sc )
{
    struct Scsi_Host *host = NULL;
    iscsi_hba_t *hba = NULL;
    iscsi_session_t *session;

    if ( ! sc ) {
        return FAILED;
    }
    host = sc->host;
    if (! host) {
        printk("iSCSI: device reset, no host for SCSI command %p\n", sc);
        return FAILED;
    }
    hba = (iscsi_hba_t *)host->hostdata;
    if (!hba) {
        printk("iSCSI: device reset, no iSCSI HBA associated with SCSI command %p\n", sc);
        return FAILED;
    }

    /* find the appropriate session for the command */
    session = find_session_for_cmnd(sc);
    if (!session) {
        printk("iSCSI: can't reset device for cmnd %p, no session\n", sc);
        return FAILED;
    }
    
    RELEASE_IO_REQUEST_LOCK;

#if 0
    /* see if we have a task for this command */
    spin_lock(&session->task_lock);
    if ((task = remove_task_for_cmnd(&session->rx_abort_tasks, sc))) {
        /* probably had an abort sent already */
        add_task(&session->tx_lun_reset_tasks, task);
    }
    else if ((task = remove_task_for_cmnd(&session->tx_abort_tasks, sc))) {
        /* waiting to have an abort sent */
        add_task(&session->tx_lun_reset_tasks, task);
    }
    else if ((task = remove_task_for_cmnd(&session->rx_tasks, sc))) {
        /* it's waiting for data, an R2T, or a command response */
        add_task(&session->tx_lun_reset_tasks, task);
    }
    else if ((task = remove_task_for_cmnd(&session->tx_tasks, sc))) {
        /* it's received an R2T, and is queued to have data sent */
        add_task(&session->tx_lun_reset_tasks, task);
    }
    else if ((task = remove_task_for_cmnd(&session->completing_tasks, sc))) {
        /* already received command completion */
        add_task(&session->tx_lun_reset_tasks, task);
    }
    spin_unlock(&session->task_lock);

    if (task) {
        /* we found a task, try a LUN reset */
        wake_tx_thread(TX_LUN_RESET, session);

        /* wait for the reset */
        timeout = jiffies + (session->AbortTimeout * HZ);

        while (jiffy_lt(jiffies, timeout) && test_bit(TX_LUN_RESET, &session->control_bits)) {
            /* log once per second indicating that we're waiting */
            if ((last_log == 0) || jiffy_lt(last_log + HZ, jiffies)) {
                last_log = jiffies;
                printk("iSCSI: waiting for LUN reset to be sent for task %p\n", task);
            }
            /* wait for a bit */
            set_current_state(TASK_INTERRUPTIBLE);
            schedule_timeout(interval);
        }

        /* to determine when the LUN reset is complete */
    }
    else 
#endif
    {
        /* without a task, just do a target reset.  It's not worth the effort to code
         * a way to queue up the LUN numbers for LUN resets with no associated task.
         */
        printk("iSCSI: eh_device_reset about to reset target for command %p for (%u %u %u %u), Cmd 0x%x\n",
               sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        if (cold_reset_target(session)) {
            printk("iSCSI: device reset succeeded for command %p, (%u %u %u %u), Cmd 0x%x\n",
                   sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
            REACQUIRE_IO_REQUEST_LOCK;
            return SUCCESS;
        }
    }

    printk("iSCSI: device reset failed for command %p, (%u %u %u %u), Cmd 0x%x\n",
           sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
    REACQUIRE_IO_REQUEST_LOCK;
    return FAILED;
}

int
iscsi_eh_bus_reset( Scsi_Cmnd *sc )
{
    struct Scsi_Host *host = NULL;
    iscsi_hba_t *hba = NULL;
    iscsi_session_t *session;

    if ( ! sc ) {
        return FAILED;
    }
    host = sc->host;
    if (! host) {
        printk("iSCSI: bus reset, no host for SCSI command %p\n", sc);
        return FAILED;
    }
    hba = (iscsi_hba_t *)host->hostdata;
    if (!hba) {
        printk("iSCSI: bus reset, no iSCSI HBA associated with SCSI command %p\n", sc);
        return FAILED;
    }

    /* find the appropriate session for the command */
    session = find_session_for_cmnd(sc);
    if (!session) {
        printk("iSCSI: can't reset device for cmnd %p, no session\n", sc);
        return FAILED;
    }
    
    RELEASE_IO_REQUEST_LOCK;

    printk("iSCSI: eh_bus_reset about to reset target for command %p for (%u %u %u %u), Cmd 0x%x\n",
           sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
    if (cold_reset_target(session)) {
        printk("iSCSI: bus reset succeeded for command %p, (%u %u %u %u), Cmd 0x%x\n",
               sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        REACQUIRE_IO_REQUEST_LOCK;
        return SUCCESS;
    }

    printk("iSCSI: bus reset failed for command %p, (%u %u %u %u), Cmd 0x%x\n",
           sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
    REACQUIRE_IO_REQUEST_LOCK;
    return FAILED;
}



int
iscsi_eh_host_reset( Scsi_Cmnd *sc )
{
    struct Scsi_Host *host = NULL;
    iscsi_hba_t *hba = NULL;
    iscsi_session_t *session;

    if ( ! sc ) {
        return FAILED;
    }
    host = sc->host;
    if (! host) {
        printk("iSCSI: host reset, no host for SCSI command %p\n", sc);
        return FAILED;
    }
    hba = (iscsi_hba_t *)host->hostdata;
    if (!hba) {
        printk("iSCSI: host reset, no iSCSI HBA associated with SCSI command %p\n", sc);
        return FAILED;
    }

    /* find the appropriate session for the command */
    session = find_session_for_cmnd(sc);
    if (!session) {
        printk("iSCSI: can't reset device for cmnd %p, no session\n", sc);
        return FAILED;
    }

    RELEASE_IO_REQUEST_LOCK;

    printk("iSCSI: eh_host_reset about to reset target for command %p for (%u %u %u %u), Cmd 0x%x\n",
           sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
    if (cold_reset_target(session)) {
        printk("iSCSI: host reset succeeded for command %p, (%u %u %u %u), Cmd 0x%x\n",
               sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        REACQUIRE_IO_REQUEST_LOCK;
        return SUCCESS;
    }

    printk("iSCSI: host reset failed for command %p, (%u %u %u %u), Cmd 0x%x\n",
           sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
    REACQUIRE_IO_REQUEST_LOCK;
    return FAILED;
}


int
iscsi_queue( Scsi_Cmnd *sc, void (*done)(Scsi_Cmnd *) )
{
    iscsi_hba_t *hba;
    iscsi_session_t *session;

    if (sc->host == NULL) {
        printk("iSCSI: queuecommand but no Scsi_Host\n");
        sc->result = HOST_BYTE(DID_NO_CONNECT);
        done(sc);
        return 0;
    }

    hba = (iscsi_hba_t *)sc->host->hostdata;
    if ( (!hba) || (!hba->active) ) {
        printk("iSCSI: queuecommand but no HBA\n");
        sc->result = HOST_BYTE(DID_NO_CONNECT);
        done(sc);
        return 0;
    }

    if ( ! TimerActive ) {
        /* iSCSI coming up or going down, fail the command */
        DEBUG_QUEUE6("iSCSI: no timer, failing to queue %p to (%u %u %u %u), Cmd 0x%x\n",
                     sc, hba->hostno, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
    
        sc->result = HOST_BYTE(DID_NO_CONNECT);
        done(sc);
        return 0;
    }
    
    if (sc->target >= ISCSI_MAX_TARGETS) {
        printk("iSCSI: invalid target id %u, failing to queue %p to (%u %u %u %u), Cmd 0x%x\n",
               sc->target, sc, hba->hostno, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        sc->result = HOST_BYTE(DID_NO_CONNECT);
        done(sc);
        return 0;
    }
    if (sc->lun >= ISCSI_MAX_LUN) {
        printk("iSCSI: invalid LUN %u, failing to queue %p to (%u %u %u %u), Cmd 0x%x\n",
               sc->lun, sc, hba->hostno, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        sc->result = HOST_BYTE(DID_NO_CONNECT );
        done(sc);
        return 0;
    }
    /* CDBs larger than 16 bytes require additional header segments, not yet implemented */
    if (sc->cmd_len > 16) {
        printk("iSCSI: cmd_len %u too large, failing to queue %p to (%u %u %u %u), Cmd 0x%x\n",
               sc->cmd_len, sc, hba->hostno, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        sc->result = HOST_BYTE(DID_NO_CONNECT );
        done(sc);
        return 0;
    }

    RELEASE_IO_REQUEST_LOCK;

#ifdef DEBUG_QUEUE
    printk("iSCSI: queueing %p to (%u %u %u %u), Cmd 0x%x, current %p\n",
           sc, hba->hostno, sc->channel, sc->target, sc->lun, sc->cmnd[0], current);
#endif

    sc->scsi_done = done;
    sc->result = 0;
    sc->host_scribble = NULL;
    ISCSI_TRACE( ISCSI_TRACE_Qd, sc->lun, sc->cmnd[0], sc->target, 0, (unsigned int)sc );
    
    if (hba) {
        cpu_flags_t cpu_flags;

        /* dispatch directly to the session's queue here, or fail.
         * this works since we keep iscsi_session_t structures around
         * even after a session drops, and re-use them if we get a new
         * session from the daemon for the same target.
         */
        spin_lock_irqsave(&hba->session_lock, cpu_flags);
        session = hba->session_list_head;
        while (session) {
            if (session->channel == sc->channel && session->target_id == sc->target) {
                /* add it to the session's command queue */
                spin_lock(&session->scsi_cmnd_lock);
                if (session->scsi_cmnd_head) {
                    /* append at the tail */
                    session->scsi_cmnd_tail->host_scribble = (unsigned char *)sc;
                    session->scsi_cmnd_tail = sc;
                }
                else {
                    /* make it the head */
                    session->scsi_cmnd_head = session->scsi_cmnd_tail = sc;
                }
                spin_unlock(&session->scsi_cmnd_lock);

                /* unlock, signal the tx thread, and return */
                spin_unlock_irqrestore(&hba->session_lock, cpu_flags);
                DEBUG_QUEUE1("iSCSI: queued %p\n", sc);
                wake_tx_thread(TX_SCSI_COMMAND, session);
                REACQUIRE_IO_REQUEST_LOCK;
                return 0;
            }
            session = session->next;
        }
        spin_unlock_irqrestore(&hba->session_lock, cpu_flags);
    }

    /* it's normal for there to be no session while the module is initializing. */
    if (test_bit(0, &init_module_complete))
        printk("iSCSI: queuecommand %p failed to find a session for HBA %p, channel %d id %d lun %d\n",
               sc, hba, sc->channel, sc->target, sc->lun);

    /* Complete with an error.  A comment say we shouldn't, but
     * it's actually ok to call done() as long as we're using
     * the eh_ style error handlers.
     */
    DEBUG_QUEUE1("iSCSI: completing %p with DID_NO_CONNECT\n", sc);
    sc->result = HOST_BYTE(DID_NO_CONNECT);

    REACQUIRE_IO_REQUEST_LOCK;
    done(sc);

    return 0;
}

int
iscsi_biosparam( Disk *disk, kdev_t dev, int geom[] )
{
    DEBUG_INIT1("iSCSI: BiosParam, capacity %d\n",disk->capacity);
    geom[0] = 64;                                   /* heads */
    geom[1] = 32;                                   /* sectors */
    geom[2] = disk->capacity / (64*32);     /* cylinders */
    return 1;
}

const char *
iscsi_info( struct Scsi_Host *sh )
{
    iscsi_hba_t *hba;
    static char buffer[256];
    char *bp;

    DEBUG_INIT0("iSCSI: Info\n");
    hba = (iscsi_hba_t *)sh->hostdata;
    if ( ! hba ) {
        return NULL;
    }

    bp = &buffer[0];
    memset( bp, 0, sizeof(buffer) );
#if defined(DRIVER_INTERNAL_VERSION) && (DRIVER_INTERNAL_VERSION > 0)        
    sprintf(bp, "iSCSI (%d.%d.%d.%d)", 
                DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION, DRIVER_PATCH_VERSION, DRIVER_INTERNAL_VERSION);
#else
    sprintf(bp, "iSCSI (%d.%d.%d)", 
                DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION, DRIVER_PATCH_VERSION);
#endif

    return bp;
}

static char *show_session_luns(iscsi_session_t *session, char *bp, int hostno)
{
    /* if we've already found LUNs, show them all */
    int lfound = 0;
    int l;

    /* FIXME: how much can we write to the buffer? */

    for (l=0; l<ISCSI_MAX_LUN; l++) {
        if (test_bit(l, session->lun_bitmap)) {
            if (session->address_length == 4) {
                bp += sprintf(bp, "  %3.1x   0 %3.1x %3.1x    %u.%u.%u.%u %5d  %s\n", 
                              hostno, session->target_id, l, 
                              session->ip_address[0], session->ip_address[1], session->ip_address[2], session->ip_address[3],
                              session->Port, session->TargetName);
            }
            else if (session->address_length == 16) {
                bp += sprintf(bp, "  %3.1x   0 %3.1x %3.1x    %02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x  %5d  %s\n", 
                              hostno, session->target_id, l, 
                              session->ip_address[0], session->ip_address[1], session->ip_address[2], session->ip_address[3],
                              session->ip_address[4], session->ip_address[5], session->ip_address[6], session->ip_address[7],
                              session->ip_address[8], session->ip_address[9], session->ip_address[10], session->ip_address[11],
                              session->ip_address[12], session->ip_address[13], session->ip_address[14], session->ip_address[15],
                              session->Port, session->TargetName);
            }
            lfound += 1;
        }
    }
    /* if we haven't found any LUNs, use ??? for a LUN number */
    if ( ! lfound ) {
        if (session->address_length == 4) {
            bp += sprintf(bp, "  %3.1x   0 %3.1x   ?    %u.%u.%u.%u %5d  %s\n", 
                          hostno, session->target_id, 
                          session->ip_address[0], session->ip_address[1], session->ip_address[2], session->ip_address[3],
                          session->Port, session->TargetName);
        }
        else if (session->address_length == 16) {
            bp += sprintf(bp, "  %3.1x   0 %3.1x   ?    %02x:%02x:%02x:%02x%02x:%02x:%02x:%02x%02x:%02x:%02x:%02x%02x:%02x:%02x:%02x  %5d  %s\n", 
                          hostno, session->target_id, 
                          session->ip_address[0], session->ip_address[1], session->ip_address[2], session->ip_address[3],
                          session->ip_address[4], session->ip_address[5], session->ip_address[6], session->ip_address[7],
                          session->ip_address[8], session->ip_address[9], session->ip_address[10], session->ip_address[11],
                          session->ip_address[12], session->ip_address[13], session->ip_address[14], session->ip_address[15],
                          session->Port, session->TargetName);
        }
    }

    return bp;
}

int iscsi_proc_info( char *buffer,
                     char **start,
                     off_t offset,
                     int length,
                     int hostno,
                     int func )
{
    char *bp = buffer;
    iscsi_hba_t *hba;
    iscsi_session_t *session;
    cpu_flags_t cpu_flags;

    bp += sprintf(bp,"# SCSI:              iSCSI:\n");
    bp += sprintf(bp,"# Hst Chn Tgt Lun    IP address     port  TargetName\n");

    /* find the HBA corresponding to hostno */
    spin_lock(&hba_list_lock);
    hba = hba_list;
    while (hba && hba->hostno != hostno)
        hba = hba->next;
    spin_unlock(&hba_list_lock);

    if (!hba) {
        printk("iSCSI: couldn't find iSCSI HBA #%d\n", hostno);
        return 0;
    }

    /* show every session on the HBA */
    spin_lock_irqsave(&hba->session_lock, cpu_flags);
    session = hba->session_list_head;
    while (session) {
        bp = show_session_luns(session, bp, hostno);
        session = session->next;
    }
    spin_unlock_irqrestore(&hba->session_lock, cpu_flags);


    *start = buffer + offset;

    if ( (bp-buffer) < offset ) {
        return 0;
    }

    if ( (bp-buffer-offset) < length ) {
        return (bp-buffer-offset);
    }

    return length;
}

/*
 * We cannot include scsi_module.c because the daemon has not got a connection
 * up yet.
 */
static int
ControlOpen( struct inode *inode, struct file *file )
{
    MOD_INC_USE_COUNT;
    return 0;
}

static int
ControlClose( struct inode *inode, struct file *file )
{
    MOD_DEC_USE_COUNT;
    return 0;
}

static int
ControlIoctl( struct inode *inode,
              struct file *file,
              unsigned int cmd,
              unsigned long arg)
{
    int rc;
    iscsi_session_ioctl_t ioctld;

    if ( ! current->files ) {
        printk("iSCSI: driver does not match kernel\n");
        return -EFAULT;
    }

    if ( cmd == ISCSI_TIMER ) {
        if ( TimerActive ) {
            return -EBUSY;
        }
        rc = iscsi_timer_thread();
        return rc;
    }

    if ( cmd == ISCSI_SESSION ) {
        if ( copy_from_user(&ioctld,(void *)arg,sizeof(iscsi_session_ioctl_t)) ) {
            DEBUG_ERR0("iSCSI: Cannot copy user Ioctl args\n");
            return -EINVAL;
        }

        rc = iscsi_session( &ioctld );

        return rc;
    }

#ifdef DEBUG_TRACE
    if ( cmd == ISCSI_GETTRACE ) {
        if (copy_to_user((void *)arg, &TraceTable[0],
                         (sizeof(iscsiTraceEntry)*ISCSI_TRACE_COUNT)))
            return -EFAULT;
        return TraceIndex;
    }
#endif

    return -EINVAL;
}

Scsi_Host_Template iscsiDriverTemplate = ISCSI;

int
init_module( void )
{
    DEBUG_INIT0("iSCSI init module\n");

#if defined(DRIVER_INTERNAL_VERSION) && (DRIVER_INTERNAL_VERSION > 0)
    /* show internal version number */
    printk("iSCSI version %d.%d.%d.%d (%s)\n",
           DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION, DRIVER_PATCH_VERSION, DRIVER_INTERNAL_VERSION,
           ISCSI_DATE);
#else
    /* released version number (INTERNAL == 0) */
    printk("iSCSI version %d.%d.%d (%s)\n",
           DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION, DRIVER_PATCH_VERSION,
           ISCSI_DATE);
#endif

#if defined(BUILD_STR)
    /* show a build string if we have one */
    if (NULL != BUILD_STR) {
        printk("iSCSI %s\n", (char *)BUILD_STR);
    }
#endif

    ControlMajor = register_chrdev( 0, ControlName, &ControlFops );
    if ( ControlMajor < 0 ) {
        printk("iSCSI failed to register the control device\n");
        return ControlMajor;
    }
    printk("iSCSI control device major number %d\n", ControlMajor);

#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
    iscsiDriverTemplate.module = THIS_MODULE;
#else
    iscsiDriverTemplate.module = &__this_module;
#endif
    scsi_register_module( MODULE_SCSI_HA, &iscsiDriverTemplate );
    if ( ! iscsiDriverTemplate.present ) {
        DEBUG_ERR0("iSCSI scsi_register not present\n");
        scsi_unregister_module( MODULE_SCSI_HA, &iscsiDriverTemplate );
    }

    DEBUG_INIT0("iSCSI init module complete\n");
    set_bit(0, &init_module_complete);
    return 0;
}

void
cleanup_module( void )
{
    int rc;

    DEBUG_INIT0("iSCSI cleanup module\n");
    if ( ControlMajor > 0 ) {
        rc = unregister_chrdev( ControlMajor, ControlName );
        if ( rc ) {
            DEBUG_ERR0("iSCSI Error trying to unregister control device\n");
        } else {
            ControlMajor = 0;
        }
    }

    if ( iscsiDriverTemplate.present ) {
        DEBUG_INIT0("iSCSI SCSI present\n");
        scsi_unregister_module( MODULE_SCSI_HA, &iscsiDriverTemplate );
        iscsiDriverTemplate.present = 0;
    }

    DEBUG_INIT0("iSCSI cleanup module complete\n");
    return;
}


/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4 
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
